import uuid
import pickle as pkl
from .adapter import get_adapter
from ._node_processor import TransformProcessor, PredictProcessor, resolve_columns
import numpy as np
import pandas as pd
import os
import shutil
import time

def _build_sub(node_attrs, data_dict, fit_process, logger):
    method = node_attrs['method']
    if method in ['transform', 'fit_transform']:
        obj = TransformProcessor(node_attrs['name'], node_attrs['processor'], node_attrs['adapter'], node_attrs['params'], logger = logger)
    else:
        obj = PredictProcessor(node_attrs['name'], node_attrs['processor'], method, node_attrs['adapter'], node_attrs['params'], logger = logger)

    # (train, train_v) only — strip valid from ((train, train_v), valid)
    fit_data = {key: val[0] for key, val in data_dict.items()}

    start_time = time.time()
    if fit_process:
        result = obj.fit_process(fit_data)
    else:
        result = None
        obj.fit(fit_data)
    elapsed_time = time.time() - start_time

    train_X, train_v_X = fit_data['X']
    info = {
        'build_id': str(uuid.uuid4()),
        'fit_time': elapsed_time,
        'train_shape': train_X.get_shape() if train_X is not None else None,
        'train_v_shape': train_v_X.get_shape() if train_v_X is not None else None
    }
    return obj, result, info

def _build_iter(node_attrs, data_dict_it, logger):
    method = node_attrs['method']
    if method in ['transform', 'predict', 'predict_proba']:
        fit_process = False
    elif method in ['fit_transform', 'fit_predict']:
        fit_process = True
    else:
        raise ValueError(f"Unknown processor_type: {method}")
    
    for data_dict in data_dict_it:
        yield _build_sub(node_attrs, data_dict, fit_process, logger), data_dict

def _build_iter_output(node_attrs, data_dict_it, logger):
    method = node_attrs['method']
    if method in ['transform', 'predict', 'predict_proba']:
        fit_process = False
    elif method in ['fit_transform', 'fit_predict']:
        fit_process = True
    else:
        raise ValueError(f"Unknown processor_type: {method}")
    
    for data_dict in data_dict_it:
        obj, result, info =  _build_sub(node_attrs, data_dict, fit_process, logger)
        (train_X, train_v_X), valid_X = data_dict['X']
        if result is None:
            train_result = obj.process(train_X)
        else:
            train_result = result
        if train_v_X is not None:
            train_v_result = obj.process(train_v_X)
        else:
            train_v_result = None
        output_train = (train_result, train_v_result)
        output_valid = obj.process(valid_X)
        yield obj, result, info, data_dict, output_train, output_valid

class StageObj():
    def __init__(self, path):
        self.path = path
        self.status = None
        self.error = None

    def load(self):
        if not os.path.isdir(self.path):
            self.status = 'finalized'
            self.objs_ = None
            return

        self.objs_ = []
        idx = 0
        while True:
            objs = []
            no = 0
            while True:
                filename = self.path / ('obj' + str(idx) + '_' + str(no) + '.pkl')
                if not os.path.isfile(filename):
                    break
                with open(filename, 'rb') as f:
                    objs.append(pkl.load(f))
                no += 1
            if len(objs) == 0:
                break
            self.objs_.append(objs)
            idx += 1

        if len(self.objs_) == 0:
            self.status = 'finalized'
            self.objs_ = None
        else:
            self.status = 'built'

    def start_exp(self, finalize = False):
        if finalize:
            raise ValueError("StageObj cannot be finalized after Experiment")

    def exp_idx(self, idx, node_attrs, data_dict_it, logger, include_input = True, include_output = True):
        if self.status == "built":
            for data_dict, (obj, train_, spec) in zip(data_dict_it, self.objs_[idx]):
                # X key로 데이터 가져오기
                (train_X, train_v_X), valid_X = data_dict['X']
                sub_result = {'spec': spec, 'object': obj}
                if include_output:
                    if train_ is None:
                        train_result = obj.process(train_X)
                    else:
                        train_result = train_
                    if train_v_X is not None:
                        train_v_result = obj.process(train_v_X)
                    else:
                        train_v_result = None
                    if include_output:
                        sub_result['output_train'] = (train_result, train_v_result)
                        sub_result['output_valid'] = obj.process(valid_X)
                    if include_input:
                        sub_result['input'] = data_dict
                    yield sub_result
        elif self.status == "finalized":
            raise RuntimeError(f"Node is finalized and cannot be re-experimented")
        else:
            raise RuntimeError(f"StageObj cannot be experimented unless built")
    
    def end_exp(self):
        pass
    
    def start_build(self):
        self.objs_ = list()
        if not os.path.isdir(self.path):
            os.makedirs(self.path, exist_ok = True)
    
    def build_idx(self, idx, node_attrs, data_dict_it, logger):
        if idx != len(self.objs_):
            raise RuntimeError(f"Build sequence is not valid")
        objs = list()
        for no, (obj, _) in enumerate(_build_iter(node_attrs, data_dict_it,logger)):
            filename = self.path / ('obj' + str(idx) + '_'  + str(no) + '.pkl')
            with open(filename, 'wb') as f:
                pkl.dump(obj, f)
            objs.append(obj)
        self.objs_.append(objs)

    def end_build(self):
        self.status = 'built'
    
    def get_objs(self, idx):
        for obj, train_, spec in self.objs_[idx]:
            yield obj, train_, spec
        
    def finalize(self):
        self.status = 'finalized'
        if os.path.isdir(self.path):
            shutil.rmtree(self.path)
        self.objs_ = None


class HeadObj():
    def __init__(self, path):
        self.path = path
        self.status = None
        self.error = None

    def load(self):
        if not os.path.isdir(self.path):
            self.status = 'finalized'
            return

        idx = 0
        while True:
            filename = self.path / ('obj' + str(idx) + '_0.pkl')
            if not os.path.isfile(filename):
                break
            idx += 1

        if idx == 0:
            self.status = 'finalized'
        else:
            self.status = 'built'

    def start_exp(self, finalize = False):
        self.finalize_after_exp = finalize
        if not os.path.isdir(self.path):
            os.makedirs(self.path, exist_ok = True)

    def exp_idx(self, idx, node_attrs, data_dict_it, logger, include_input = True, include_output = True):
        if self.status == "built":
            no = 0
            for data_dict in data_dict_it:
                filename = self.path / ('obj' + str(idx) + '_'  + str(no) + '.pkl')
                if not os.path.isfile(filename):
                    break
                with open(filename, 'rb') as f:
                    obj, train_, spec = pkl.load(f)
                # X key로 데이터 가져오기
                (train_X, train_v_X), valid_X = data_dict['X']
                sub_result = {'spec': spec, 'object': obj}
                if include_output:
                    if train_ is None:
                        train_result = obj.process(train_X)
                    else:
                        train_result = train_
                    if train_v_X is not None:
                        train_v_result = obj.process(train_v_X)
                    else:
                        train_v_result = None
                    if include_output:
                        sub_result['output_train'] = (train_result, train_v_result)
                        sub_result['output_valid'] = obj.process(valid_X)
                    if include_input:
                        sub_result['input'] = data_dict
                    yield sub_result
                no += 1
        elif self.status == "finalized":
            raise RuntimeError(f"Node is finalized and cannot be re-experimented")
        else:
            no = 0
            if include_output:
                objs_iter = _build_iter_output(node_attrs, data_dict_it, logger)
                for obj, train_, spec, data_dict, output_train, output_valid in objs_iter:
                    sub_result = {'spec': spec, 'object': obj}
                    sub_result['output_train'] = output_train
                    sub_result['output_valid'] = output_valid
                    if include_input:
                        sub_result['input'] = data_dict
                    yield sub_result
                    if self.status is None and (not self.finalize_after_exp):
                        filename = self.path / ('obj' + str(idx) + '_'  + str(no) + '.pkl')
                        with open(filename, 'wb') as f:
                            pkl.dump((obj, train_, spec), f)
                    no += 1
            else:
                objs_iter = _build_iter(node_attrs, data_dict_it, logger)
                for obj, train_, spec, data_dict in objs_iter:
                    sub_result = {'spec': spec, 'object': obj}
                    if include_input:
                        sub_result['input'] = data_dict
                    yield sub_result
                    if self.status is None and (not self.finalize_after_exp):
                        filename = self.path / ('obj' + str(idx) + '_'  + str(no) + '.pkl')
                        with open(filename, 'wb') as f:
                            pkl.dump((obj, train_, spec), f)
                    no += 1
    
    def end_exp(self):
        if self.finalize_after_exp:
            self.status = "finalized"
        else:
            self.status = "built"

    def get_objs(self, idx):
        if self.status != 'built':
            raise RuntimeError(f"must be built before accessing objects (status='{self.status}')")
        no = 0
        while True:
            filename = self.path / ('obj' + str(idx) + '_'  + str(no) + '.pkl')
            if not os.path.isfile(filename):
                break
            with open(filename, 'rb') as f:
                obj, train_, spec = pkl.load(f)
            yield obj, train_, spec
            no += 1

    def finalize(self):
        self.status = 'finalized'
        if os.path.isdir(self.path):
            shutil.rmtree(self.path)



import re
import os
import sys
import uuid
import pickle as pkl
import shutil
import traceback
import warnings
from pathlib import Path

import pandas as pd
from cachetools import LRUCache

from sklearn.model_selection import ShuffleSplit

from ._data_wrapper import wrap, unwrap
from ._expobj import HeadObj, StageObj
from ._describer import desc_spec, desc_status, desc_obj_vars
from ._logger import DefaultLogger

from ._pipeline import Pipeline
from ._node_processor import resolve_columns
from ._connector import Connector
from .collector import Collector, MetricCollector, StackingCollector, ModelAttrCollector, SHAPCollector, OutputCollector
from ._trainer import Trainer

def _get_data_size(data):
    if data is None:
        return 0
    if isinstance(data, list):
        return sum(_get_data_size(item) for item in data)
    if isinstance(data, tuple):
        return sum(_get_data_size(item) for item in data)
    if hasattr(data, 'nbytes'):
        return data.nbytes
    if hasattr(data, 'memory_usage'):
        return data.memory_usage(deep=True).sum()
    return sys.getsizeof(data)

class DataCache():
    def __init__(self, maxsize=4 * 1024 ** 3):  # 4GB 기본값
        self.cache_dic = LRUCache(maxsize=maxsize, getsizeof=_get_data_size)

    def get_data(self, node, typ, idx):
        key = (node, typ, idx)
        return self.cache_dic.get(key, None)

    def put_data(self, node, typ, idx, data):
        key = (node, typ, idx)
        self.cache_dic[key] = data

    def clear(self):
        self.cache_dic.clear()

    def clear_nodes(self, nodes):
        node_set = set(nodes)
        keys_to_delete = [k for k in self.cache_dic.keys() if k[0] in node_set]
        for k in keys_to_delete:
            del self.cache_dic[k]

class Experimenter():
    def __init__(
            self, data, path, data_names = None, sp = ShuffleSplit(n_splits=1, random_state=1), sp_v=None,
            splitter_params=None, title=None, data_key=None, cache_maxsize=4 * 1024 ** 3,
            logger = DefaultLogger(level=['info', 'progress'])
        ):
        self.cache_maxsize = cache_maxsize
        self.logger = logger
        self.path = Path(path)
        if not os.path.exists(path):
            self.path.mkdir(parents=True, exist_ok=True)
            self.logger.info(f"📁 Created directory: {self.path}")
        self.train_idx_list = list()
        self.valid_idx_list = list()
        data_native = data
        self.data = wrap(data)
        # 실험 타이틀 저장
        self.title = title

        # data 식별자 (load 시 검증용)
        self.data_key = data_key

        # splitter 설정 저장
        self.sp = sp
        self.sp_v = sp_v
        self.splitter_params = splitter_params if splitter_params is not None else {}
        self.exp_id = str(uuid.uuid4())

        split_params = {}

        if data_names is None:
            data_names = self.data.get_columns()
        for k, v in self.splitter_params.items():
            split_params[k] = unwrap(self.data.select_columns(v))

        for train_idx, valid_idx in sp.split(data_native, **split_params):
            if sp_v is not None:
                train_data = self.data.iloc(train_idx)
                train_data_native = unwrap(train_data)

                inner_split_params = {'X': train_data_native}
                for k, v in self.splitter_params.items():
                    inner_split_params[k] = unwrap(train_data.select_columns(v))

                self.train_idx_list.append([
                    (train_idx[train_v_idx], train_idx[valid_v_idx])
                    for train_v_idx, valid_v_idx in sp_v.split(**inner_split_params)
                ])
            else:
                self.train_idx_list.append([
                    (train_idx, None)
                ])
            self.valid_idx_list.append(valid_idx)

        self.pipeline = Pipeline()
        self.node_objs = {}
        self.cache = DataCache(maxsize=cache_maxsize)
        self.grps = {}
        self.collectors = {}
        self.trainers = {}
        self.status = "open"
        self._save()

    def _check_open(self):
        """상태가 open인지 확인하고, 아니면 에러 발생"""
        if self.status != "open":
            raise RuntimeError(f"Experimenter is '{self.status}'. Only 'open' status allows modifications.")

    def open(self):
        """Experimenter를 open 상태로 변경"""
        self.status = "open"
        self._save()
        self.logger.info("Experimenter status changed to 'open'")

    def close(self):
        """Experimenter를 close 상태로 변경"""
        self.status = "close"
        self._save()
        self.logger.info("Experimenter status changed to 'close'")

    @staticmethod
    def create(data, path, data_names=None, sp=ShuffleSplit(n_splits=1, random_state=1), sp_v=None,
            splitter_params=None, title=None, data_key=None, cache_maxsize=4 * 1024 ** 3,
            logger = DefaultLogger(level=['info', 'progress'])):

        if os.path.exists(path):
            raise RuntimeError(f"Exists: {path}")
        return Experimenter(
            data, path, data_names, sp=sp, sp_v=sp_v, splitter_params=splitter_params,
            title=title, data_key=data_key, cache_maxsize=cache_maxsize, logger=logger)

    def get_n_splits(self):
        return len(self.train_idx_list)

    def get_n_splits_inner(self):
        return len(self.train_idx_list[0])

    def add_collector(self, collector, exist = 'skip'):
        if collector.name in self.collectors:
            if exist == 'skip':
                return self.collectors[collector.name]
            elif exist == 'error':
                raise RuntimeError("")
        
        self._check_open()
        collector.path = self.path / '__collector' / collector.name
        collector.save()
        self.collectors[collector.name] = collector
        self.collect(collector)
        self._save()
        return collector

    def add_trainer(self, name, data=None, splitter="same", splitter_params=None, exist='skip'):
        if name in self.trainers:
            if exist == 'skip':
                return self.trainers[name]
            elif exist == 'error':
                raise RuntimeError(f"Trainer '{name}' already exists")

        if data is None:
            trainer_data = self.data
        else:
            trainer_data = wrap(data)

        if splitter == 'same':
            if splitter_params is not None:
                raise ValueError("splitter_params must be None when splitter='same'")
            trainer_splitter = self.sp_v
            trainer_splitter_params = self.splitter_params
        else:
            trainer_splitter = splitter
            trainer_splitter_params = splitter_params if splitter_params is not None else {}

        trainer = Trainer(
            name=name,
            pipeline=self.pipeline,
            data=trainer_data,
            path=self.path / '__trainer' / name,
            splitter=trainer_splitter,
            splitter_params=trainer_splitter_params,
            logger=self.logger,
            cache=self.cache
        )
        self.trainers[name] = trainer
        self._save()
        return trainer

    def _validate_name(self, name):
        """Node 또는 NodeGroup 이름 검증

        Args:
            name: 검증할 이름

        Raises:
            ValueError: 이름이 유효하지 않을 경우
        """
        if name is None:
            return

        # '__' 포함 금지
        if '__' in name:
            raise ValueError(f"Name '{name}' cannot contain '__'")

        # 파일/폴더명으로 사용 불가한 문자 금지
        invalid_chars = ['/', '\\', '\0', '<', '>', ':', '"', '|', '?', '*']
        for char in invalid_chars:
            if char in name:
                raise ValueError(f"Name '{name}' cannot contain '{char}'")

    def get_grp_path(self, grp):
        if self.path is None:
            return None
        if isinstance(grp, str):
            grp = self.pipeline.get_grp(grp)
        path_parts = [grp.name]
        current = self.pipeline.get_grp(grp.parent)
        while current is not None:
            path_parts.insert(0, current.name)
            current = self.pipeline.get_grp(current.parent)
        return self.path / '/'.join(path_parts)

    def get_node_path(self, node):
        if isinstance(node, str):
            node = self.pipeline.get_node(node)
        grp_path = self.get_grp_path(node.grp)
        return grp_path / node.name

    def set_grp(self, name, role=None, processor=None, edges=None, method=None, parent=None, adapter=None, params=None, exist = 'skip'):
        self._check_open()
        result_obj = self.pipeline.set_grp(
            name, role, processor, edges, method, parent, adapter, params, exist
        )
        
        affected_nodes = result_obj['affected_nodes']
        self.reset_nodes(affected_nodes)
        new_grp_path = self.get_grp_path(result_obj['grp'])
        if "old_grp" in result_obj:
            old_grp_path = self.get_grp_path(result_obj['old_grp'])
            if old_grp_path != new_grp_path:
                os.makedirs(new_grp_path, exist_ok=True)
                for fname in os.listdir(old_grp_path):
                    src_path = os.path.join(old_grp_path, fname)
                    dst_path = os.path.join(new_grp_path, fname)
                    shutil.move(src_path, dst_path)

            self.logger.info(f"Group '{name}' updated, {len(affected_nodes)} node(s) affected")
        self._save()
        return result_obj

    def rename_grp(self, name_from, name_to):
        self._check_open()
        old_grp_path = self.get_grp_path(name_from)
        self.pipeline.rename_grp(name_from, name_to)
        new_grp_path = self.get_grp_path(name_to)
        if old_grp_path.exists():
            os.makedirs(new_grp_path, exist_ok=True)
            for fname in os.listdir(old_grp_path):
                src_path = os.path.join(old_grp_path, fname)
                dst_path = os.path.join(new_grp_path, fname)
                shutil.move(src_path, dst_path)
            shutil.rmtree(old_grp_path)
        self._save()
    
    def remove_grp(self, name):
        self._check_open()
        self.pipeline.remove_grp(name)

        self.logger.info(f"Group '{name}' removed")
        self._save()


    def remove_node(self, name):
        """노드를 제거

        Args:
            name: 제거할 노드 이름

        Raises:
            ValueError: 노드가 존재하지 않거나, 자식 노드가 있는 경우
        """
        self._check_open()
        self.pipeline.remove_node(name)
        for v in self.collectors.values():
            v.reset_nodes([name])

        self.logger.info(f"Node '{name}' removed")
        self._save()

    def finalize(self, nodes):
        self._check_open()
        node_names = self.pipeline.get_node_names(nodes)
        for i in node_names:
            if i is None:
                continue
            node = self.pipeline.get_node(i)
            grp = self.pipeline.get_grp(node.grp)
            if grp.role == 'head' and i in self.node_objs:
                node_obj = self.node_objs[i]
                if node_obj.status == 'built':
                    self.logger.info(f"Finalize '{i}'")
                    node_obj.finalize()

    def reinitialize(self, nodes):
        self._check_open()
        node_names = self.pipeline.get_node_names(nodes)
        for i in node_names:
            if i is None:
                continue
            if i in self.node_objs:
                node_obj = self.node_objs[i]
                if node_obj.status == 'finalized':
                    self.logger.info(f"reinitialize '{i}'")
                    del self.node_objs[i]

    def close_exp(self):
        if self.status != "open":
            raise RuntimeError("")
        for k, node_obj in self.node_objs.items():
            if node_obj.status == 'built':
                self.logger.info(f"Finalize '{k}'")
                node_obj.finalize()
        self.status = "closed"
    
    def reopen_exp(self):
        if self.status != "closed":
            raise RuntimeError("")
        for k in list(self.node_objs.keys()):
            node = self.pipeline.get_node(k)
            grp = self.pipeline.get_grp(node.grp)
            if grp.role == 'stage':
                self.logger.info(f"Initialize '{k}'")
                del self.node_objs[k]
        self.build()

    def set_node(
        self, name, grp, processor = None, edges = None,
        method = None, adapter = None, params = None, exist = 'skip'
    ):
        self._check_open()
        result_obj = self.pipeline.set_node(
            name, grp, processor, edges, method, adapter, params, exist
        )

        # 기존 노드를 업데이트한 경우, 하위 노드들도 재빌드
        if len(result_obj['affected_nodes']) > 0:
            affected_nodes = result_obj['affected_nodes']
            self.logger.info(f"Affected {len(affected_nodes)} dependent node(s): {sorted(affected_nodes)}")
            self.reset_nodes(affected_nodes)

        if result_obj['result'] == 'update':
            self.reset_nodes([name])
        self._save()
        return result_obj

    def reset_nodes(self, nodes):
        for i in nodes:
            if i in self.node_objs:
                del self.node_objs[i]

        self.cache.clear_nodes(nodes)

        for v in self.collectors.values():
            v.reset_nodes(nodes)

        for v in self.trainers.values():
            v.reset_nodes(nodes)

    def build(self, nodes = None, rebuild = False):
        self._check_open()
        node_names = self.pipeline.get_node_names(nodes)
        target_nodes = list()
        for i in self.pipeline._get_affected_nodes([None]):
            node = self.pipeline.get_node(i)
            grp = self.pipeline.get_grp(node.grp)
            if grp.role == 'stage' and i not in self.node_objs:
                target_nodes.append(i)

        self.logger.info(f"Building {len(target_nodes)} node(s)")
        for node in target_nodes:
            if node in self.node_objs:
                node_obj = self.node_objs[node]
            else:
                node_obj = StageObj(self.get_node_path(node))
                self.node_objs[node] = node_obj
            node_obj.start_build()
        
        n_splits = self.get_n_splits()
        self.logger.start_progress("Build", n_splits)
        for i in range(n_splits):
            self.logger.update_progress(i)
            self.logger.start_progress("Node", len(target_nodes))
            for ni, node in enumerate(target_nodes):
                self.logger.update_progress(ni)
                self.logger._progress[-1][0] = node
                node_obj = self.node_objs[node]
                if node_obj.status == 'error':
                    continue
                try:
                    with warnings.catch_warnings(record=True) as caught:
                        warnings.simplefilter("always")
                        node_attrs = self.pipeline.get_node_attrs(node)
                        node_obj.build_idx(
                            i, node_attrs, self.get_node_data(node, i), self.logger
                        )
                        for w in caught:
                            self.logger.warning(f"[{node}] fold {i}: {w.category.__name__}: {w.message}")
                except Exception as e:
                    node_obj.status = 'error'
                    node_obj.error = {
                        'type': type(e).__name__,
                        'message': str(e),
                        'traceback': traceback.format_exc(),
                        'fold': i,
                    }
                    self.logger.info(f"[{node}] Build error at fold {i}: {type(e).__name__}: {e}")
                    self.logger.info(traceback.format_exc())
            self.logger.end_progress(len(target_nodes))
        self.logger.end_progress(n_splits)

        error_nodes = [n for n in target_nodes if self.node_objs[n].status == 'error']
        for node in target_nodes:
            node_obj = self.node_objs[node]
            if node_obj.status != 'error':
                node_obj.end_build()
        if error_nodes:
            self.logger.info(f"Build complete: {len(target_nodes) - len(error_nodes)}/{len(target_nodes)} node(s), {len(error_nodes)} error(s): {error_nodes}")
        else:
            self.logger.info(f"Build complete: {len(target_nodes)} node(s)")
    
    def exp(self, nodes = None):
        self._check_open()
        node_names = set(self.pipeline.get_node_names(nodes))
        target_nodes = list()
        for i in self.pipeline._get_affected_nodes([None]):
            node = self.pipeline.get_node(i)
            grp = self.pipeline.get_grp(node.grp)
            if grp.role == 'head' and i not in self.node_objs and i in node_names:
                target_nodes.append(i)

        self.logger.info(f"Experimenting {len(target_nodes)} node(s)")

        # connector matching
        node_attrs_cache = {n: self.pipeline.get_node_attrs(n) for n in target_nodes}
        matched = {}
        for name, collector in self.collectors.items():
            matched[name] = set(
                n for n in target_nodes
                if collector.connector.match(n, node_attrs_cache[n])
            )

        # start_experiment for all nodes
        for node in target_nodes:
            if node in self.node_objs:
                node_obj = self.node_objs[node]
            else:
                node_obj = HeadObj(self.get_node_path(node))
                self.node_objs[node] = node_obj
            node_obj.start_exp()

        # collector _start
        for name, collector in self.collectors.items():
            for node in target_nodes:
                if node in matched[name]:
                    collector._start(node)

        # experiment loop
        n_splits = self.get_n_splits()
        self.logger.start_progress("Exp", n_splits)
        for i in range(n_splits):
            self.logger.update_progress(i)

            self.logger.start_progress("Node", len(target_nodes))
            for ni, node in enumerate(target_nodes):
                self.logger.update_progress(ni)
                self.logger._progress[-1][0] = node
                node_obj = self.node_objs[node]
                if node_obj.status == 'error':
                    continue
                try:
                    with warnings.catch_warnings(record=True) as caught:
                        warnings.simplefilter("always")
                        node_attrs = node_attrs_cache[node]
                        result_iter = node_obj.exp_idx(
                            i, node_attrs, self.get_node_data(node, i), self.logger
                        )

                        for inner_idx, result_data in enumerate(result_iter):
                            context = {
                                'node_attrs': node_attrs,
                                'processor': result_data['object'],
                                'spec': result_data['spec'],
                                'input': result_data['input'],
                                'output_train': result_data['output_train'],
                                'output_valid': result_data['output_valid'],
                            }
                            for name, collector in self.collectors.items():
                                if node in matched[name]:
                                    collector._collect(node, i, inner_idx, context)

                        for w in caught:
                            self.logger.warning(f"[{node}] fold {i}: {w.category.__name__}: {w.message}")

                    # collector _end_idx
                    for name, collector in self.collectors.items():
                        if node in matched[name]:
                            collector._end_idx(node, i)
                except Exception as e:
                    node_obj.status = 'error'
                    node_obj.error = {
                        'type': type(e).__name__,
                        'message': str(e),
                        'traceback': traceback.format_exc(),
                        'fold': i,
                    }
                    self.logger.info(f"[{node}] Exp error at fold {i}: {type(e).__name__}: {e}")
                    self.logger.info(traceback.format_exc())
                    for name, collector in self.collectors.items():
                        if node in matched[name]:
                            collector.reset_nodes([node])

            self.logger.end_progress(len(target_nodes))
        self.logger.end_progress(n_splits)

        error_nodes = [n for n in target_nodes if self.node_objs[n].status == 'error']
        # end_experiment for non-error nodes
        for node in target_nodes:
            node_obj = self.node_objs[node]
            if node_obj.status != 'error':
                node_obj.end_exp()

        # collector _end for non-error nodes
        for name, collector in self.collectors.items():
            for node in target_nodes:
                if node in matched[name] and self.node_objs[node].status != 'error':
                    collector._end(node)

        if error_nodes:
            self.logger.info(f"Experimentation complete: {len(target_nodes) - len(error_nodes)}/{len(target_nodes)} node(s), {len(error_nodes)} error(s): {error_nodes}")
        else:
            self.logger.info(f"Experimentation complete: {len(target_nodes)} node(s)")
        self._save()

    def collect(self, collector, exist = 'skip'):
        # built head 노드 중 connector 매칭
        target_nodes = []
        node_attrs_cache = {}
        for name in self.pipeline._get_affected_nodes([None]):
            node = self.pipeline.get_node(name)
            grp = self.pipeline.get_grp(node.grp)
            if name not in self.node_objs:
                continue
            if exist == 'skip' and collector.has(name):
                continue
            node_obj = self.node_objs[name]
            if node_obj.status != 'built':
                continue
            node_attrs = self.pipeline.get_node_attrs(name)
            if collector.connector.match(name, node_attrs) and not collector.has_node(name):
                target_nodes.append(name)
                node_attrs_cache[name] = node_attrs

        for node in target_nodes:
            collector._start(node)

        n_splits = self.get_n_splits()
        self.logger.start_progress("Collect", n_splits)
        for idx in range(n_splits):
            self.logger.update_progress(idx)
            self.logger.start_progress("Node", len(target_nodes))
            for ni, node in enumerate(target_nodes):
                self.logger.update_progress(ni)
                self.logger._progress[-1][0] = node
                node_obj = self.node_objs[node]
                node_attrs = node_attrs_cache[node]
                result_iter = node_obj.exp_idx(
                    idx, node_attrs, self.get_node_data(node, idx), self.logger
                )
                for inner_idx, result_data in enumerate(result_iter):
                    context = {
                        'node_attrs': node_attrs,
                        'processor': result_data['object'],
                        'spec': result_data['spec'],
                        'input': result_data['input'],
                        'output_train': result_data['output_train'],
                        'output_valid': result_data['output_valid'],
                    }
                    collector._collect(node, idx, inner_idx, context)
                collector._end_idx(node, idx)
            self.logger.end_progress(len(target_nodes))
        self.logger.end_progress(n_splits)

        for node in target_nodes:
            collector._end(node)

        return collector

    def get_data(self, idx, edges):
        data_dict = {}
        for key, edge_list in edges.items():
            key_data_list = []
            for node_name, var in edge_list:
                key_data_list.append(self.get_node_output(idx, node_name, var))
            data_dict[key] = zip(*key_data_list)

        for _ in range(self.get_n_splits_inner()):
            result = {}
            for k, it in data_dict.items():
                z = next(it)
                train_sub, valid_sub, outer_valid_sub = list(), list(), list()
                for (train_data, train_v_data), outer_valid_data in z:
                    train_sub.append(train_data)
                    if train_v_data is not None:
                        valid_sub.append(train_v_data)
                    outer_valid_sub.append(outer_valid_data)

                train_concat = type(train_sub[0]).concat(train_sub, axis=1)
                outer_concat = type(outer_valid_sub[0]).concat(outer_valid_sub, axis=1)
                if len(valid_sub) > 0:
                    valid_concat = type(valid_sub[0]).concat(valid_sub, axis=1)
                    result[k] = ((train_concat, valid_concat), outer_concat)
                else:
                    result[k] = ((train_concat, None), outer_concat)
            yield result
    
    def get_data_train(self, idx, edges):
        data_dict = {}
        for key, edge_list in edges.items():
            key_data_list = []
            for node_name, var in edge_list:
                key_data_list.append(self.get_node_train_output(idx, node_name, var))
            data_dict[key] = zip(*key_data_list)
        
        for _ in range(self.get_n_splits_inner()):
            result = {}
            for k, it in data_dict.items():
                z = next(it)
                train_sub, valid_sub = list(), list()
                for train_data, train_v_data in z:
                    train_sub.append(train_data)
                    if train_v_data is not None:
                        valid_sub.append(train_v_data)
                train_concat = type(train_sub[0]).concat(train_sub, axis=1)
                if len(valid_sub) > 0:
                    valid_concat = type(valid_sub[0]).concat(valid_sub, axis=1)
                    result[k] = (train_concat, valid_concat)
                else:
                    result[k] = (train_concat, None)
            yield result
    
    def get_data_valid(self, idx, edges):
        data_dict = {}
        for key, edge_list in edges.items():
            key_data_list = []
            for node_name, var in edge_list:
                key_data_list.append(self.get_node_valid_output(idx, node_name, var))
            data_dict[key] = zip(*key_data_list)
        for _ in range(self.get_n_splits_inner()):
            result = {}
            for k, it in data_dict.items():
                z = next(it)
                outer_valid_sub = list()
                for outer_valid_data in z:
                    outer_valid_sub.append(outer_valid_data)
                outer_concat = type(outer_valid_sub[0]).concat(outer_valid_sub, axis=1)
                result[k] = outer_concat
            yield result

    def get_node_data(self, node, idx):
        node_attrs = self.pipeline.get_node_attrs(node)
        return self.get_data(idx, node_attrs['edges'])
    
    def get_node_data_train(self, node, idx):
        node_attrs = self.pipeline.get_node_attrs(node)
        return self.get_data_train(idx, node_attrs['edges'])

    def get_node_data_valid(self, node, idx):
        node_attrs = self.pipeline.get_node_attrs(node)
        return self.get_data_valid(idx, node_attrs['edges'])

    def split(self, edges):
        for idx in range(len(self.train_idx_list)):
            yield self.get_data(idx, edges)
    
    def get_node_output(self, idx, node, v = None):
        if node is None:
            outer_valid_data = self.data.iloc(self.valid_idx_list[idx])
            if v is not None:
                outer_valid_data = outer_valid_data.select_columns(v)
            
            for train_v_idx, valid_v_idx in self.train_idx_list[idx]:
                if v is None:
                    train_data = self.data.iloc(train_v_idx)
                    train_v_data = self.data.iloc(valid_v_idx) if valid_v_idx is not None else None
                else:
                    train_data = self.data.iloc(train_v_idx).select_columns(v)
                    if valid_v_idx is not None:
                        train_v_data = self.data.iloc(valid_v_idx).select_columns(v)
                    else:
                        train_v_data = None

                yield (train_data, train_v_data), outer_valid_data
            return

        cached = self.cache.get_data(node, "all", idx)
        if cached is not None:
            sub = self.node_objs[node].get_objs(idx)
            for ((train_result, train_v_result), valid_result), (obj, _, _) in zip(cached, sub):
                X = resolve_columns(train_result, v, processor=obj)
                train_result = train_result.select_columns(X)
                if train_v_result is not None:
                    train_v_result = train_v_result.select_columns(X)
                valid_result = valid_result.select_columns(X)
                yield (train_result, train_v_result), valid_result
            return
        it = self.get_node_data(node, idx)
        sub = self.node_objs[node].get_objs(idx)
        use_cache = self.cache_maxsize > 0
        cache_data = list() if use_cache else None
        
        for data_dict, (obj, train_, info) in zip(it, sub):
            # X key로 입력 데이터 가져오기
            (train_X, train_v_X), valid_X = data_dict['X']

            # train data 처리
            if train_ is None:
                train_result = obj.process(train_X)
            else:
                train_result = train_

            # train_v data 처리
            if train_v_X is not None:
                train_v_result = obj.process(train_v_X)
            else:
                train_v_result = None

            # valid data 처리 (외부 fold의 valid)
            valid_result = obj.process(valid_X)
            if use_cache:
                cache_data.append(((train_result, train_v_result), valid_result))
            # 필요하면 컬럼 필터링
            if v is not None:
                X = resolve_columns(train_result, v, processor=obj)
                train_result = train_result.select_columns(X)
                if train_v_result is not None:
                    train_v_result = train_v_result.select_columns(X)
                valid_result = valid_result.select_columns(X)

            yield (train_result, train_v_result), valid_result
        if use_cache:
            self.cache.put_data(node, "all", idx, cache_data)

    def get_node_train_output(self, idx, node, v=None):
        if node is None:
            for train_v_idx, valid_v_idx in self.train_idx_list[idx]:
                if v is None:
                    train_data = self.data.iloc(train_v_idx)
                    train_v_data = self.data.iloc(valid_v_idx) if valid_v_idx is not None else None
                else:
                    train_data = self.data.iloc(train_v_idx).select_columns(v)
                    if valid_v_idx is not None:
                        train_v_data = self.data.iloc(valid_v_idx).select_columns(v)
                    else:
                        train_v_data = None

                yield train_data, train_v_data
            return
        cached = self.cache.get_data(node, "train", idx)
        if cached is not None:
            sub = self.node_objs[node].get_objs(idx)
            for (train_result, train_v_result), (obj, _, _) in zip(cached, sub):
                X = resolve_columns(train_result, v, processor=obj)
                train_result = train_result.select_columns(X)
                if train_v_result is not None:
                    train_v_result = train_v_result.select_columns(X)
                yield train_result, train_v_result
            return

        it = self.get_node_data_train(node, idx)
        sub = self.node_objs[node].get_objs(idx)
        cache_data = list()
        for data_dict, (obj, train_, info) in zip(it, sub):
            train_X, train_v_X = data_dict['X']
            train_result = obj.process(train_X) if train_ is None else train_
            if train_v_X is not None:
                train_v_result = obj.process(train_v_X)
            else:
                train_v_result = None
            cache_data.append((train_result, train_v_result))
            if v is not None:
                X = resolve_columns(train_result, v, processor=obj)
                train_result = train_result.select_columns(X)
                if train_v_result is not None:
                    train_v_result = train_v_result.select_columns(X)
            yld = train_result, train_v_result
            yield yld
        self.cache.put_data(node, "train", idx, cache_data)
    
    def get_node_valid_output(self, idx, node, v=None):
        if node is None:
            outer_valid_data = self.data.iloc(self.valid_idx_list[idx])
            if v is not None:
                outer_valid_data = outer_valid_data.select_columns(v)
            
            for _ in range(self.get_n_splits_inner()):
                yield outer_valid_data

        cached = self.cache.get_data(node, "valid", idx)
        if cached is not None:
            sub = self.node_objs[node].get_objs(idx)
            for valid_result, (obj, _, _) in zip(cached, sub):
                X = resolve_columns(valid_result, v, processor=obj)
                valid_result = valid_result.select_columns(X)
                yield valid_result

        it = self.get_node_data_valid(node, idx)
        sub = self.node_objs[node].get_objs(idx)
        use_cache = self.cache_maxsize > 0
        cache_data = list() if use_cache else None
        for data_dict, (obj, train_, info) in zip(it, sub):
            # X key로 valid 데이터 가져오기
            valid_X = data_dict['X']
            # valid data 처리 (외부 fold의 valid)
            valid_result = obj.process(valid_X)
            if use_cache:
                cache_data.append(valid_result)
            # 필요하면 컬럼 필터링
            if v is not None:
                X = resolve_columns(valid_result, v, processor=obj)
                valid_result = valid_result.select_columns(X)
            yield valid_result
        if use_cache:
            self.cache.put_data(node, "valid", idx, cache_data)
        return ret_func()
    
    def get_node_info(self):
        lines = [f"# Experiment Pipeline Summary\n"]
        lines.append(f"- **DataSource**\n")

        for name in self.pipeline.nodes.keys():
            if name is None:
                continue
            node = self.pipeline.get_node(name)
            node_attrs = node.get_attrs(self.pipeline.grps)
            processor_name = node_attrs['processor'].__name__ if node_attrs['processor'] else 'None'
            edges_info_parts = []
            for key, edge_list in node_attrs['edges'].items():
                edge_strs = [f"{n or 'DataSource'}{f'[{v}]' if v else ''}" for n, v in edge_list]
                edges_info_parts.append(f"{key}: [{', '.join(edge_strs)}]")
            edges_info = ", ".join(edges_info_parts)
            lines.append(f"## {name}")
            lines.append(f"- **Processor**: {processor_name}")
            lines.append(f"- **Method**: {node_attrs['method']}")
            lines.append(f"- **Edges**: {edges_info}")

            descendants = self.pipeline._find_descendants(name)
            if descendants:
                lines.append(f"- **Descendants**: {sorted(descendants)}")
            lines.append("")

        return "\n".join(lines)

    def get_objs(self, node_name, idx):
        if node_name not in self.node_objs or node_name is None:
            raise ValueError(f"Node '{node_name}' objects not found")

        obj = self.node_objs[node_name]

        # 노드가 빌드되지 않았으면 에러
        if obj.status != 'built':
            raise ValueError(f"Node '{node_name}' status should be built")

        # 외부 fold의 내부 fold들: [(processor, train_v, info), ...]
        return obj.get_objs(idx)
    
    def get_obj_vars(self, node_name, idx):
        if node_name not in self.node_objs:
            raise ValueError(f"Node '{node_name}' has no object")

        # 외부 fold의 내부 fold들: [(processor, train_v, info), ...]
        objs_ = self.node_objs[node_name].get_objs(idx)

        # (입력변수 튜플, 출력변수 튜플) -> 내부 fold index 리스트
        var_map = {}
        for inner_idx, (processor, train_v, info) in enumerate(objs_):
            # 입력 변수와 출력 변수 가져오기
            input_vars = tuple(processor.X_) if hasattr(processor, 'X_') and processor.X_ is not None else ()
            output_vars = tuple(processor.output_vars) if hasattr(processor, 'output_vars') and processor.output_vars is not None else ()

            # 튜플 키 생성
            key = (input_vars, output_vars)

            if key not in var_map:
                var_map[key] = []
            var_map[key].append(inner_idx)

        # 결과 리스트 생성: [(입력변수 리스트, 출력변수 리스트, 내부 폴드 index 리스트), ...]
        result = []
        for (input_vars, output_vars), fold_indices in var_map.items():
            result.append((list(input_vars), list(output_vars), fold_indices))

        # 등장 빈도(내부 폴드 개수)의 내림차순으로 정렬
        result.sort(key=lambda x: len(x[2]), reverse=True)
        return result

    def get_edges_var(self, edges):
        class _ColHolder:
            def __init__(self, columns):
                self._columns = columns
            def get_columns(self):
                return self._columns

        var_map = {}

        for idx in range(self.get_n_splits()):
            n_inner = len(self.train_idx_list[idx])
            # edges는 dict: {key: [(node_name, var), ...], ...}
            edge_objs = {}
            for key, edge_list in edges.items():
                edge_objs[key] = []
                for node_name, var in edge_list:
                    if node_name is None:
                        edge_objs[key].append((None, var, None))
                    else:
                        node_obj = self.node_objs[node_name]
                        edge_objs[key].append((node_name, var, node_obj.get_objs(idx)))

            for inner_idx in range(n_inner):
                collected = {}
                for key, objs_list in edge_objs.items():
                    collected[key] = []
                    for node_name, var, objs in objs_list:
                        if node_name is None:
                            cols = self.data.get_columns()
                            proc = None
                        else:
                            obj = next(objs)
                            proc = obj[0]
                            cols = list(proc.output_vars) if proc.output_vars is not None else []

                        if var is not None:
                            cols = resolve_columns(_ColHolder(cols), var, processor=proc)

                        collected[key].extend(cols)

                # key를 정렬된 형태로 튜플화
                key_tuple = tuple((k, tuple(v)) for k, v in sorted(collected.items()))
                if key_tuple not in var_map:
                    var_map[key_tuple] = []
                var_map[key_tuple].append((idx, inner_idx))

        result = []
        for vars_tuple, fold_indices in var_map.items():
            # dict로 복원
            vars_dict = {k: list(v) for k, v in vars_tuple}
            result.append((vars_dict, fold_indices))

        result.sort(key=lambda x: len(x[1]), reverse=True)

        return result

    def _save(self, filepath=None):
        if filepath is None:
            filepath = self.path / '__exp.pkl'

        save_data = {
            'data_key': self.data_key,
            'title': self.title,
            'sp': self.sp,
            'sp_v': self.sp_v,
            'splitter_params': self.splitter_params,
            'cache_maxsize': self.cache_maxsize,
            'exp_id': self.exp_id,
            'pipeline': self.pipeline,
            'node_obj_keys': list(self.node_objs.keys()),
            'collector_keys': {name: type(c).__name__ for name, c in self.collectors.items()},
            'trainer_keys': list(self.trainers.keys()),
            'status': self.status
        }

        with open(filepath, 'wb') as f:
            pkl.dump(save_data, f)

    @staticmethod
    def load(filepath, data, data_key=None):
        COLLECTOR_TYPES = {
            'MetricCollector': MetricCollector,
            'StackingCollector': StackingCollector,
            'ModelAttrCollector': ModelAttrCollector,
            'SHAPCollector': SHAPCollector,
        }

        filepath = Path(filepath)
        with open(filepath / '__exp.pkl', 'rb') as f:
            save_data = pkl.load(f)

        saved_data_key = save_data.get('data_key')
        if saved_data_key is not None and saved_data_key != data_key:
            raise ValueError(
                f"data_key mismatch: saved='{saved_data_key}', provided='{data_key}'"
            )

        exp = Experimenter(
            data=data,
            path=filepath,
            sp=save_data['sp'],
            sp_v=save_data['sp_v'],
            splitter_params=save_data['splitter_params'],
            title=save_data['title'],
            data_key=saved_data_key,
            cache_maxsize=save_data.get('cache_maxsize', 4 * 1024 ** 3)
        )
        exp.exp_id = save_data['exp_id']
        exp.pipeline = save_data['pipeline']
        exp.status = save_data['status']

        # node_objs 복원
        for node_name in save_data['node_obj_keys']:
            node = exp.pipeline.get_node(node_name)
            grp = exp.pipeline.get_grp(node.grp)
            node_path = exp.get_node_path(node_name)

            if grp.role == 'stage':
                node_obj = StageObj(node_path)
            else:
                node_obj = HeadObj(node_path)
            node_obj.load()
            exp.node_objs[node_name] = node_obj

        # Collector 복원
        collector_keys = save_data.get('collector_keys', {})
        for coll_name, type_name in collector_keys.items():
            cls = COLLECTOR_TYPES.get(type_name)
            if cls is None:
                continue
            coll_path = filepath / '__collector' / coll_name
            if (coll_path / '__config.pkl').exists():
                collector = cls.load(coll_path)
                exp.collectors[coll_name] = collector

        # Trainer 복원
        from ._trainer import Trainer
        for trainer_name in save_data.get('trainer_keys', []):
            trainer_path = filepath / '__trainer' / trainer_name
            if (trainer_path / '__trainer.pkl').exists():
                trainer = Trainer._load(
                    trainer_path,
                    pipeline=exp.pipeline,
                    data=exp.data,
                    cache=exp.cache,
                    logger=exp.logger,
                )
                exp.trainers[trainer_name] = trainer

        exp.logger.info(f"Loaded: {len(exp.pipeline.nodes) - 1} node(s), {len(exp.pipeline.grps)} group(s), {len(exp.train_idx_list)} fold(s)")
        return exp
    
    def export_pipeline(self):
        return self.pipeline.copy()

    def import_pipeline(self, pipeline):
        if len(self.pipeline.nodes) > 0 or len(self.pipeline.grps) > 0:
            raise RuntimeError("")
        self.pipeline = pipeline.copy()
    
    def desc_status(self):
        return desc_status(self)

    def desc_spec(self):
        return desc_spec(self)

    def desc_obj_vars(self, node_name, idx):
        obj_vars = self.get_obj_vars(node_name, idx)
        return desc_obj_vars(self, obj_vars[0])

    def desc_pipeline(self, max_depth=None, direction='TD'):
        return self.pipeline.desc_pipeline(max_depth, direction)

    def desc_node(self, node_name, direction='TD', show_params=False):
        return self.pipeline.desc_node(node_name, direction, show_params)


import numpy as np

from ._base import DataFilter


class RandomFilter(DataFilter):
    def __init__(self, n=None, frac=None, random_state=None):
        if n is not None and frac is not None:
            raise ValueError("n and frac are mutually exclusive")
        self.n = n
        self.frac = frac
        self.random_state = random_state

    def _select(self, data_dict):
        first_val = next(iter(data_dict.values()))
        total = first_val.get_shape()[0]

        if self.n is not None:
            sample_n = min(self.n, total)
        elif self.frac is not None:
            sample_n = int(total * self.frac)
        else:
            return np.arange(total)

        rng = np.random.RandomState(self.random_state)
        return np.sort(rng.choice(total, size=sample_n, replace=False))



"""
CatBoost adapter
"""

import tempfile
import json
import pandas as pd
from ._base import ModelAdapter


class CatBoostAdapter(ModelAdapter):
    """Adapter for CatBoost models (CatBoostClassifier, CatBoostRegressor)

    CatBoost도 eval_set을 지원합니다.
    """

    def get_fit_params(self, data_dict, X, y=None, params=None, logger=None):
        """CatBoost의 fit 파라미터 구성"""
        from .._data_wrapper import unwrap

        fit_params = {}

        # data_dict에서 데이터 추출
        train_X, train_v_X = data_dict['X']
        if 'y' in data_dict:
            train_y, train_v_y = data_dict['y']
        else:
            train_y, train_v_y = None, None

        # eval_set 구성
        if self.eval_mode and self.eval_mode != 'none' and train_v_X is not None and train_v_y is not None:
            train_X_native = unwrap(train_X)
            train_y_native = unwrap(train_y)
            train_v_X_native = unwrap(train_v_X)
            train_v_y_native = unwrap(train_v_y)

            if self.eval_mode == 'valid':
                fit_params['eval_set'] = [(train_v_X_native, train_v_y_native)]
            elif self.eval_mode == 'both':
                fit_params['eval_set'] = [(train_X_native, train_y_native), (train_v_X_native, train_v_y_native)]

        # verbose 처리
        if self.verbose > 0:
            if self.verbose < 1:
                # 0 < verbose < 1: 진행률 기반 출력
                # CatBoost는 복잡한 callback 구조라서 간단히 기본 verbose 사용
                fit_params['verbose'] = False
            else:
                # verbose >= 1: CatBoost 기본 verbose (iteration 단위)
                fit_params['verbose'] = int(self.verbose)
        else:
            # verbose == 0: 출력 안함
            fit_params['verbose'] = False

        return fit_params

    @staticmethod
    def _get_feature_importances_pvc(processor):
        obj = processor.obj
        input_vars = list(processor.X_) if hasattr(processor, 'X_') and processor.X_ is not None else list(range(obj.feature_count_))

        return pd.Series(
            obj.get_feature_importance(type='PredictionValuesChange'),
            index=input_vars, name = 'PredictionValuesChange'
        )

    @staticmethod
    def _get_feature_importances_interaction(processor):
        obj = processor.obj
        input_vars = list(processor.X_) if hasattr(processor, 'X_') and processor.X_ is not None else list(range(obj.feature_count_))

        interaction = obj.get_feature_importance(type='Interaction')
        return pd.DataFrame(
            interaction, columns=['feat1', 'feat2', 'importance']
        ).assign(
            feat1=lambda x: x['feat1'].astype('int').apply(lambda y: input_vars[y]),
            feat2=lambda x: x['feat2'].astype('int').apply(lambda y: input_vars[y]),
        ).set_index(['feat1', 'feat2'])['importance']

    @staticmethod
    def _get_evals_result(processor):
        obj = processor.obj
        evals_result = obj.get_evals_result() if hasattr(obj, 'get_evals_result') else {}
        return pd.concat(
            [pd.DataFrame(v).stack().rename(k) for k, v in evals_result.items()], axis=1
        ).stack()

    @staticmethod
    def _get_trees(processor):
        obj = processor.obj
        with tempfile.NamedTemporaryFile(suffix=".json") as f:
            obj.save_model(f.name, format="json")
            trees = json.load(f).get('oblivious_trees', [])
        return trees

CatBoostAdapter.result_objs = {
    'feature_importances_pvc': (CatBoostAdapter._get_feature_importances_pvc, True),
    'feature_importances_interaction': (CatBoostAdapter._get_feature_importances_interaction, True),
    'evals_result': (CatBoostAdapter._get_evals_result, True), 
    'trees': (CatBoostAdapter._get_trees, False)
}


from sklearn.base import BaseEstimator, TransformerMixin

class PandasConverter(TransformerMixin, BaseEstimator):
    def __init__(self, index_col = None):
        self.index_col = index_col
        self.columns_ = None

    def fit(self, X, y = None):
        self.columns_ = [i for i in X.columns if i != self.index_col]
        self.fitted_ = True
        return self
    
    def transform(self, X):
        df = X.to_pandas()
        if self.index_col is not None and self.index_col in df.columns:
            df = df.set_index(self.index_col)
        return df

    def get_params(self, deep = True):
        return {
            'index_col': self.index_col
        }

    def set_output(self, transform = 'pandas'):
        pass

    def get_feature_names_out(self, X = None):
        return self.columns_


import pickle
import pandas as pd

from ._base import Collector
from ..adapter import get_adapter


class ModelAttrCollector(Collector):
    def __init__(self, name, connector, result_key, adapter=None, params=None):
        super().__init__(name, connector)
        self.result_key = result_key
        if adapter is not None:
            self.adapter = adapter
        else:
            self.adapter = get_adapter(connector.processor)
            if self.result_key not in self.adapter.result_objs:
                raise RuntimeError("")
        self.params = params or {}
        self.results = {}
        self._sub = {}

    def _start(self, node):
        self.results[node] = []
        self._sub[node] = []

    def _collect(self, node, idx, inner_idx, context):
        result_func = self.adapter.result_objs[self.result_key][0]
        result = result_func(context['processor'], **self.params)
        self._sub[node].append(result)

    def _end_idx(self, node, idx):
        self.results[node].append(self._sub[node])
        self._sub[node] = []

    def _end(self, node):
        if len(self.results[node]) == 0:
            del self.results[node]
        if node in self._sub:
            del self._sub[node]
        self.save()

    def has_node(self, node):
        return node in self.results

    def reset_nodes(self, nodes):
        for node in nodes:
            if node in self.results:
                del self.results[node]
        self.save()

    def save(self):
        if self.path is None:
            return
        self._ensure_path()
        data = {
            'name': self.name,
            'connector': self.connector,
            'result_key': self.result_key,
            'adapter': self.adapter,
            'params': self.params,
            'results': self.results,
        }
        with open(self.path / '__config.pkl', 'wb') as f:
            pickle.dump(data, f)

    @classmethod
    def load(cls, path):
        with open(path / '__config.pkl', 'rb') as f:
            data = pickle.load(f)
        obj = cls(
            name=data['name'],
            connector=data['connector'],
            result_key=data['result_key'],
            adapter=data.get('adapter'),
            params=data['params'],
        )
        obj.results = data['results']
        obj.path = path
        return obj

    def _is_mergeable(self):
        if self.adapter is None or self.result_key not in self.adapter.result_objs:
            return False
        return self.adapter.result_objs[self.result_key][1]

    def get_attr(self, node, idx=None):
        if idx is not None:
            return self.results[node][idx]
        return self.results[node]

    def get_attrs(self, nodes=None):
        node_names = self._get_nodes(nodes, self.results.keys())
        return {node: self.results[node] for node in node_names}

    def get_attrs_agg(self, node, agg_inner=True, agg_outer=True):
        if agg_outer and not agg_inner:
            raise ValueError("agg_outer requires agg_inner to be True")
        if not self._is_mergeable():
            raise ValueError(f"Result '{self.result_key}' is not mergeable across folds")
        l = list()
        for no, inner_list in enumerate(self.results[node]):
            l.append(
                pd.concat([j.rename(no_i) for no_i, j in enumerate(inner_list)], axis=1).stack().rename(no)
            )
        df = pd.concat(l, axis=1)
        if agg_inner:
            df = df.groupby(level=[i for i in range(len(df.index.levels) - 1)]).mean()
            if agg_outer:
                return df.mean(axis=1)
        return df




from sklearn.base import BaseEstimator, TransformerMixin

class LGBMIterativeImputer(TransformerMixin, BaseEstimator):
    def __init__(self, hparams, X_num, X_cat, targets, na_value = None, hparams_dic = {}, na_value_dic = {}, 
                 n_iter = 2, validation_fraction = 0, progress_callback=None):
        self.hparams = hparams
        self.X_num = X_num
        self.X_cat = X_cat
        self.targets = targets
        self.na_value = na_value
        self.hparams_dic = hparams_dic
        self.na_value_dic = na_value_dic
        self.n_iter = n_iter
        self.validation_fraction = validation_fraction
        self.models_ = None
        self.hist_ = None
        self.initial_cat_values_ = None
        self.progress_callback = progress_callback

    def fit(self, X, y = None):
        self.models_ = list()
        self.hist_ = list()
        self.progress_callback.on_train_begin()
        X_ = X.copy()
        if len(self.X_cat) > 0:
            self.initial_cat_values_ = X_[self.X_cat].apply(lambda x: x.mode().iloc[0])
            X_ = dproc.join_and_assign(
                X_, X_[self.X_cat].fillna(self.initial_cat_values_)
            )
        for i in range(self.n_iter):
            self.progress_callback.on_iter_begin(self.n_iter)
            self._partial_fit(X_, X)
            self.progress_callback.on_iter_end(i)
        self.progress_callback.on_train_end()
        del X_
        self.fitted_ = True
        return self

    def _partial_fit(self, X, X_org):
        if self.models_ is None:
            self.models_ = list()
            self.hist_ = list()
        if len(self.models_) >= self.n_iter:
            return self
        round_1 = list()
        round_hist = list()
        self.progress_callback.on_step_begin(len(self.targets))
        for target in self.targets:
            val = [i for i in self.X_num + self.X_cat if i != target]
            hparams = self.hparams_dic.get(target, self.hparams)
            na_value = self.na_value_dic.get(target, self.na_value)
            s_tgt_notna = X_org[target].notna() if na_value is None else X_org[target] != na_value
            if target in self.X_cat:
                round_1.append(
                    lgb.LGBMClassifier(verbose = - 1, **hparams)
                )
                if self.validation_fraction <= 0:
                    X_train = X.loc[s_tgt_notna]
                    X_test = None
                else:
                    X_train, X_test = X.loc[s_tgt_notna].pipe(
                        lambda x: train_test_split(x, test_size = self.validation_fraction, random_state = 123, stratify = x[target])
                    )
            else:
                round_1.append(
                    lgb.LGBMRegressor(verbose = -1, **hparams)
                )
                if self.validation_fraction <= 0:
                    X_train = X.loc[s_tgt_notna]
                    X_test = None
                else:
                    X_train, X_test = X.loc[s_tgt_notna].pipe(
                        lambda x: train_test_split(x, test_size = self.validation_fraction, random_state = 123)
                    )
            round_1[-1].fit(X_train[val], X_train[target], categorical_feature = [i for i in val if i in self.X_cat])
            X.loc[~s_tgt_notna, target] = X.loc[~s_tgt_notna, val].pipe(lambda x: pd.Series(round_1[-1].predict(x), index = x.index, dtype = X[target].dtype))
            if X_test is not None:
                if target in self.X_cat:
                    round_hist.append(
                        accuracy_score(X_test[target], round_1[-1].predict(X_test[val]))
                    )
                else:
                    round_hist.append(
                        mean_squared_error(X_test[target], round_1[-1].predict(X_test[val]))
                    )
            self.progress_callback.on_step_end(len(self.targets))
        self.models_.append(round_1)
        self.hist_.append(round_hist)
        return self
    
    def transform(self, X, n_iter = None):
        if n_iter is None:
            n_iter = len(self.models_)
        X_ = X.copy()
        if len(self.X_cat) > 0:
            self.initial_cat_values_ = X_[self.X_cat].apply(lambda x: x.mode().iloc[0])
            X_ = dproc.join_and_assign(
                X_, X_[self.X_cat].fillna(self.initial_cat_values_)
            )
        for i, round_1 in enumerate(self.models_):
            if i >= n_iter:
                break
            for target, m in zip(self.targets, round_1):
                na_value = self.na_value_dic.get(target, self.na_value)
                s_tgt_na = X[target].isna() if na_value is None else X[target] == na_value
                na_value = self.na_value_dic.get(target, self.na_value)
                val = [i for i in self.X_num + self.X_cat if i != target]
                X_.loc[s_tgt_na, target] = X_.loc[s_tgt_na, val].pipe(lambda x: pd.Series(m.predict(x), index = x.index, dtype = X[target].dtype))
        return X_[self.targets]

    def get_params(self, deep=True):
        return {
            'hparams':  hparams,
            'X_num': X_num,
            'X_cat': X_cat,
            'targets': targets,
            'na_value': na_value,
            'hparams_dic': hparams_dic,
            'na_value_dic': na_value_dic,
            'n_iter': n_iter,
            'validate_fraction': validate_fraction
        }
    
    def set_output(self, transform='pandas'):
        pass

    def get_feature_names_out(self, X = None):
        return [target]


import os
import time
import uuid
import pickle as pkl

from ._node_processor import TransformProcessor, PredictProcessor


def _train_build(node_attrs, data_dict, logger):
    method = node_attrs['method']
    if method in ['transform', 'fit_transform']:
        fit_process = method == 'fit_transform'
        obj = TransformProcessor(
            node_attrs['name'], node_attrs['processor'],
            node_attrs['adapter'], node_attrs['params'], logger=logger,
        )
    elif method in ['predict', 'predict_proba', 'fit_predict']:
        fit_process = method == 'fit_predict'
        obj = PredictProcessor(
            node_attrs['name'], node_attrs['processor'],
            method, node_attrs['adapter'], node_attrs['params'], logger=logger,
        )
    else:
        raise ValueError(f"Unknown method: {method}")

    start_time = time.time()
    if fit_process:
        result = obj.fit_process(data_dict)
    else:
        obj.fit(data_dict)
        result = None
    elapsed_time = time.time() - start_time

    train_X, train_v_X = data_dict['X']
    info = {
        'build_id': str(uuid.uuid4()),
        'fit_time': elapsed_time,
        'train_shape': train_X.get_shape() if train_X is not None else None,
        'train_v_shape': train_v_X.get_shape() if train_v_X is not None else None,
    }
    return obj, result, info


class TrainStageObj:

    def __init__(self, path):
        self.path = path
        self.status = None
        self.error = None
        self.objs_ = None

    def load(self):
        if not os.path.isdir(self.path):
            self.objs_ = None
            return
        self.objs_ = {}
        split_idx = 0
        while True:
            filename = self.path / f'obj{split_idx}.pkl'
            if not os.path.isfile(filename):
                break
            with open(filename, 'rb') as f:
                self.objs_[split_idx] = pkl.load(f)
            split_idx += 1
        if self.objs_:
            self.status = 'built'

    def start_build(self):
        self.objs_ = {}
        os.makedirs(self.path, exist_ok=True)

    def build_split(self, split_idx, node_attrs, data_dict, logger):
        obj, result, info = _train_build(node_attrs, data_dict, logger)
        filename = self.path / f'obj{split_idx}.pkl'
        with open(filename, 'wb') as f:
            pkl.dump((obj, result, info), f)
        self.objs_[split_idx] = (obj, result, info)

    def end_build(self):
        self.status = 'built'

    def get_obj(self):
        for i in self.objs_:
            yield self.objs_[i]


class TrainHeadObj:

    def __init__(self, path):
        self.path = path
        self.status = None
        self.error = None

    def load(self):
        if not os.path.isdir(self.path):
            return
        if os.path.isfile(self.path / 'obj0.pkl'):
            self.status = 'built'

    def start_build(self):
        os.makedirs(self.path, exist_ok=True)

    def build_split(self, split_idx, node_attrs, data_dict, logger):
        obj, result, info = _train_build(node_attrs, data_dict, logger)
        filename = self.path / f'obj{split_idx}.pkl'
        with open(filename, 'wb') as f:
            pkl.dump((obj, result, info), f)

    def end_build(self):
        self.status = 'built'

    def get_obj(self):
        no = 0
        while True:
            filename = self.path / f'obj{no}.pkl'
            if not os.path.isfile(filename):
                break
            with open(filename, 'rb') as f:
                yield pkl.load(f)
            no += 1



import re


class Connector:
    def __init__(self, node_query=None, edges=None, processor=None):
        self.node_query = node_query
        self.edges = edges
        self.processor = processor

    def match(self, node_name, node_attrs):
        if self.node_query is not None:
            if isinstance(self.node_query, str):
                if not re.search(self.node_query, node_name):
                    return False
            elif isinstance(self.node_query, list):
                if node_name not in self.node_query:
                    return False

        if self.processor is not None:
            if node_attrs.get('processor') != self.processor:
                return False

        if self.edges is not None:
            node_edges = node_attrs.get('edges', {})
            for key, required_edges in self.edges.items():
                if key not in node_edges:
                    return False
                for edge in required_edges:
                    if edge not in node_edges[key]:
                        return False

        return True



from sklearn.base import BaseEstimator, TransformerMixin
import polars as pl
from ._dproc import get_type_df, get_type_pl, merge_type_df

class PolarsLoader(TransformerMixin, BaseEstimator):
    def __init__(self, predefined_types = {}, read_method = 'read_csv'):
        self.predefined_types = predefined_types
        self.read_method = read_method
    
    def fit(self, X, y = None):
        if type(X) is list:
            self.df_type_ = merge_type_df([
                pl.scan_csv(i).pipe(get_type_df) for i in X
            ])
        else:
            self.df_type_ = pl.scan_csv(X).pipe(get_type_df)
        self.pl_type_ = get_type_pl(
            self.df_type_, self.predefined_types
        )
        self.fitted_ = True
        return self

    def transform(self, X):
        if type(X) is list:
            return pl.concat([
                getattr(pl, self.read_method)(i, schema_overrides = self.pl_type_) for i in X
            ])
        else:
            return getattr(pl, self.read_method)(X, schema_overrides = self.pl_type_)

    def get_params(self, deep = True):
        return {
            'predefined_types': self.predefined_types,
            'read_method': self.read_method
        }
        
    def set_output(self, transform = 'polars'):
        pass

    def get_feature_names_out(self, X = None):
        return [i for i in self.pl_type_.keys()]


class ExprProcessor(TransformerMixin, BaseEstimator):
    def __init__(self, dict_expr, with_columns = True):
        self.dict_expr = dict_expr
        self.with_columns = with_columns
        self.columns = []

    def fit(self, X, y = None):
        if self.with_columns:
            self.columns = X.columns
        self.fitted_ = True
        return self
    
    def transform(self, X):
        if self.with_columns:
            return X.with_columns(**{
                k:v for k, v in self.dict_expr.items() if k in X.columns
            })
        else:
            return X.select(**self.dict_expr)
    
    def get_params(self, deep=True):
        return {
            'dict_expr': self.dict_expr,
            'with_columns': self.with_columns
        }

    def set_output(self, transform = 'polars'):
        pass

    def get_feature_names_out(self, X = None):
        return self.columns + list(self.dict_expr.keys())


"""
Keras adapter
"""

from ._base import ModelAdapter


class KerasAdapter(ModelAdapter):
    """Adapter for Keras models (KerasClassifier, KerasRegressor)

    Keras는 validation_data 파라미터로 (X, y) 튜플을 받습니다.
    """

    def get_fit_params(self, data_dict, X, y=None, params=None, logger=None):
        """Keras의 fit 파라미터 구성"""
        from .._data_wrapper import unwrap

        fit_params = {}

        # data_dict에서 데이터 추출
        train_X, train_v_X = data_dict[X]
        if y is not None and y in data_dict:
            train_y, train_v_y = data_dict[y]
        else:
            train_y, train_v_y = None, None

        # validation_data 구성
        if self.eval_mode and self.eval_mode != 'none' and train_v_X is not None and train_v_y is not None:
            train_v_X_native = unwrap(train_v_X)
            train_v_y_native = unwrap(train_v_y)
            # Keras는 'valid'와 'both' 모두 동일하게 처리 (validation_data만 지원)
            fit_params['validation_data'] = (train_v_X_native, train_v_y_native)

        # verbose 처리
        if self.verbose > 0:
            if self.verbose < 1:
                # 0 < verbose < 1: Keras의 verbose는 0, 1, 2만 지원하므로 1 사용
                fit_params['verbose'] = 1
            else:
                # verbose >= 1: Keras verbose (0: silent, 1: progress bar, 2: one line per epoch)
                fit_params['verbose'] = min(int(self.verbose), 2)
        else:
            # verbose == 0: 출력 안함
            fit_params['verbose'] = 0

        return fit_params



import numpy as np


class DataFilter:
    def __call__(self, data_dict):
        indices = self._select(data_dict)
        return {key: val.iloc(indices) for key, val in data_dict.items()}

    def _select(self, data_dict):
        raise NotImplementedError



import pytest
import numpy as np
import pandas as pd
from pathlib import Path

from sklearn.preprocessing import StandardScaler
from sklearn.tree import DecisionTreeClassifier
from sklearn.model_selection import ShuffleSplit, KFold

from mllabs._experimenter import Experimenter
from mllabs import Connector, MetricCollector, StackingCollector, ModelAttrCollector, OutputCollector


def accuracy_metric(y, pred):
    return (y.values == pred.values).mean()


def dummy_metric(y, pred):
    return 0.5


@pytest.fixture
def sample_data():
    np.random.seed(42)
    n = 100
    return pd.DataFrame({
        'f1': np.random.randn(n),
        'f2': np.random.randn(n),
        'f3': np.random.randn(n),
        'target': np.random.randint(0, 2, n),
    })


@pytest.fixture
def built_exp(tmp_path, sample_data):
    e = Experimenter(
        data=sample_data,
        path=tmp_path / 'exp',
        sp=ShuffleSplit(n_splits=2, test_size=0.2, random_state=42),
    )
    e.set_grp('scale', role='stage', processor=StandardScaler,
              method='transform', edges={'X': [(None, ['f1', 'f2', 'f3'])]})
    e.set_node('scaler', grp='scale')
    e.set_grp('model', role='head', processor=DecisionTreeClassifier,
              method='predict',
              edges={'X': [('scaler', None)], 'y': [(None, 'target')]},
              params={'max_depth': 3, 'random_state': 42})
    e.set_node('dt', grp='model')
    e.build()
    e.exp()
    return e


@pytest.fixture
def built_exp_inner(tmp_path, sample_data):
    e = Experimenter(
        data=sample_data,
        path=tmp_path / 'exp_inner',
        sp=ShuffleSplit(n_splits=2, test_size=0.2, random_state=42),
        sp_v=KFold(n_splits=3, shuffle=True, random_state=42),
    )
    e.set_grp('model', role='head', processor=DecisionTreeClassifier,
              method='predict',
              edges={'X': [(None, ['f1', 'f2', 'f3'])], 'y': [(None, 'target')]},
              params={'max_depth': 3, 'random_state': 42})
    e.set_node('dt', grp='model')
    e.build()
    e.exp()
    return e


@pytest.fixture
def multi_head_exp(tmp_path, sample_data):
    e = Experimenter(
        data=sample_data,
        path=tmp_path / 'exp_multi',
        sp=ShuffleSplit(n_splits=2, test_size=0.2, random_state=42),
    )
    e.set_grp('model', role='head', processor=DecisionTreeClassifier,
              method='predict',
              edges={'X': [(None, ['f1', 'f2', 'f3'])], 'y': [(None, 'target')]},
              params={'max_depth': 3, 'random_state': 42})
    e.set_node('dt1', grp='model')
    e.set_node('dt2', grp='model', params={'max_depth': 5})
    e.build()
    e.exp()
    return e


class TestConnector:
    def test_match_all(self):
        c = Connector()
        assert c.match('any_node', {}) is True

    def test_match_node_query_str(self):
        c = Connector(node_query='dt')
        assert c.match('dt1', {}) is True
        assert c.match('scaler', {}) is False

    def test_match_node_query_regex(self):
        c = Connector(node_query='^dt')
        assert c.match('dt1', {}) is True
        assert c.match('my_dt', {}) is False

    def test_match_node_query_list(self):
        c = Connector(node_query=['dt1', 'dt2'])
        assert c.match('dt1', {}) is True
        assert c.match('dt3', {}) is False

    def test_match_processor(self):
        c = Connector(processor=DecisionTreeClassifier)
        assert c.match('dt', {'processor': DecisionTreeClassifier}) is True
        assert c.match('dt', {'processor': StandardScaler}) is False

    def test_match_edges(self):
        edges_req = {'X': [(None, ['f1'])]}
        c = Connector(edges=edges_req)
        node_attrs = {'edges': {'X': [(None, ['f1']), ('s', None)], 'y': [(None, 'target')]}}
        assert c.match('dt', node_attrs) is True

    def test_match_edges_missing_key(self):
        c = Connector(edges={'z': [(None, None)]})
        assert c.match('dt', {'edges': {'X': [(None, None)]}}) is False

    def test_match_combined(self):
        c = Connector(node_query='dt', processor=DecisionTreeClassifier)
        assert c.match('dt1', {'processor': DecisionTreeClassifier}) is True
        assert c.match('dt1', {'processor': StandardScaler}) is False
        assert c.match('scaler', {'processor': DecisionTreeClassifier}) is False


class TestMetricCollector:
    def test_collect_basic(self, built_exp):
        mc = MetricCollector('acc', Connector(), output_var=None,
                             metric_func=accuracy_metric)
        built_exp.add_collector(mc)
        assert mc.has('dt')

    def test_get_metric(self, built_exp):
        mc = MetricCollector('acc', Connector(), output_var=None,
                             metric_func=accuracy_metric)
        built_exp.add_collector(mc)
        result = mc.get_metric('dt')
        assert isinstance(result, pd.Series)
        assert result.name == 'dt'
        assert len(result) > 0
        assert all(0 <= v <= 1 for v in result.values)

    def test_get_metrics(self, multi_head_exp):
        mc = MetricCollector('acc', Connector(), output_var=None,
                             metric_func=accuracy_metric)
        multi_head_exp.add_collector(mc)
        result = mc.get_metrics()
        assert isinstance(result, pd.DataFrame)
        assert 'dt1' in result.index.get_level_values(0)
        assert 'dt2' in result.index.get_level_values(0)

    def test_get_metrics_with_node_filter(self, multi_head_exp):
        mc = MetricCollector('acc', Connector(), output_var=None,
                             metric_func=accuracy_metric)
        multi_head_exp.add_collector(mc)
        result = mc.get_metrics(nodes=['dt1'])
        assert 'dt1' in result.index.get_level_values(0)
        assert 'dt2' not in result.index.get_level_values(0)

    def test_get_metrics_regex(self, multi_head_exp):
        mc = MetricCollector('acc', Connector(), output_var=None,
                             metric_func=accuracy_metric)
        multi_head_exp.add_collector(mc)
        result = mc.get_metrics(nodes='dt1')
        assert len(result) > 0

    def test_get_metrics_agg(self, built_exp):
        mc = MetricCollector('acc', Connector(), output_var=None,
                             metric_func=accuracy_metric)
        built_exp.add_collector(mc)
        mean, std = mc.get_metrics_agg()
        assert isinstance(mean, pd.DataFrame)
        assert std is None

    def test_get_metrics_agg_with_std(self, built_exp):
        mc = MetricCollector('acc', Connector(), output_var=None,
                             metric_func=accuracy_metric)
        built_exp.add_collector(mc)
        mean, std = mc.get_metrics_agg(include_std=True)
        assert isinstance(mean, pd.DataFrame)
        assert isinstance(std, pd.DataFrame)

    def test_get_metrics_agg_inner_only(self, built_exp_inner):
        mc = MetricCollector('acc', Connector(), output_var=None,
                             metric_func=accuracy_metric)
        built_exp_inner.add_collector(mc)
        mean, std = mc.get_metrics_agg(inner_fold=True, outer_fold=False)
        assert isinstance(mean, pd.DataFrame)

    def test_get_metrics_agg_no_fold(self, built_exp):
        mc = MetricCollector('acc', Connector(), output_var=None,
                             metric_func=accuracy_metric)
        built_exp.add_collector(mc)
        result = mc.get_metrics_agg(inner_fold=False, outer_fold=False)
        assert isinstance(result, pd.DataFrame)

    def test_get_metrics_agg_invalid(self, built_exp):
        mc = MetricCollector('acc', Connector(), output_var=None,
                             metric_func=accuracy_metric)
        built_exp.add_collector(mc)
        with pytest.raises(ValueError):
            mc.get_metrics_agg(inner_fold=False, outer_fold=True)

    def test_include_train(self, built_exp):
        mc = MetricCollector('acc_train', Connector(), output_var=None,
                             metric_func=accuracy_metric, include_train=True)
        built_exp.add_collector(mc)
        result = mc.get_metric('dt')
        assert 'train_sub' in result.index.get_level_values(-1)

    def test_inner_split_metrics(self, built_exp_inner):
        mc = MetricCollector('acc', Connector(), output_var=None,
                             metric_func=accuracy_metric)
        built_exp_inner.add_collector(mc)
        result = mc.get_metric('dt')
        assert len(result) > 2

    def test_connector_filter(self, multi_head_exp):
        mc = MetricCollector('acc', Connector(node_query=['dt1']),
                             output_var=None, metric_func=accuracy_metric)
        multi_head_exp.add_collector(mc)
        assert mc.has('dt1')
        assert not mc.has('dt2')

    def test_reset_nodes(self, built_exp):
        mc = MetricCollector('acc', Connector(), output_var=None,
                             metric_func=accuracy_metric)
        built_exp.add_collector(mc)
        assert mc.has('dt')
        mc.reset_nodes(['dt'])
        assert not mc.has('dt')

    def test_save_load(self, built_exp):
        mc = MetricCollector('acc', Connector(), output_var=None,
                             metric_func=accuracy_metric)
        built_exp.add_collector(mc)
        loaded = MetricCollector.load(mc.path)
        assert loaded.has('dt')
        result_orig = mc.get_metric('dt')
        result_loaded = loaded.get_metric('dt')
        pd.testing.assert_series_equal(result_orig, result_loaded)

    def test_ad_hoc_collect(self, built_exp):
        mc = MetricCollector('acc', Connector(), output_var=None,
                             metric_func=accuracy_metric)
        built_exp.add_collector(mc)
        mc2 = MetricCollector('acc2', Connector(), output_var=None,
                              metric_func=dummy_metric)
        built_exp.add_collector(mc2)
        assert mc2.has('dt')
        result = mc2.get_metric('dt')
        assert all(v == 0.5 for v in result.values)


class TestStackingCollector:
    def test_collect_basic(self, built_exp):
        sc = StackingCollector('stk', Connector(), output_var=None,
                               experimenter=built_exp)
        built_exp.add_collector(sc)
        assert sc.has_node('dt')

    def test_get_dataset(self, built_exp):
        sc = StackingCollector('stk', Connector(
            edges={'y': [(None, 'target')]}
        ), output_var=None, experimenter=built_exp)
        built_exp.add_collector(sc)
        ds = sc.get_dataset()
        assert isinstance(ds, pd.DataFrame)
        assert len(ds) == int(len(built_exp.data.data) * 0.2) * 2 # ShuffleSplit, n_splits = 2, test_size = 0.2
        assert 'target' in ds.columns

    def test_get_dataset_no_target(self, built_exp):
        sc = StackingCollector('stk', Connector(), output_var=None,
                               experimenter=built_exp)
        built_exp.add_collector(sc)
        ds = sc.get_dataset(include_target=False)
        assert isinstance(ds, pd.DataFrame)
        assert 'target' not in ds.columns

    def test_get_dataset_multi_nodes(self, multi_head_exp):
        sc = StackingCollector('stk', Connector(
            edges={'y': [(None, 'target')]}
        ), output_var=None, experimenter=multi_head_exp)
        multi_head_exp.add_collector(sc)
        ds = sc.get_dataset()
        assert ds.shape[1] > 2

    def test_get_dataset_node_filter(self, multi_head_exp):
        sc = StackingCollector('stk', Connector(
            edges={'y': [(None, 'target')]}
        ), output_var=None, experimenter=multi_head_exp)
        multi_head_exp.add_collector(sc)
        ds = sc.get_dataset(nodes=['dt1'])
        assert isinstance(ds, pd.DataFrame)

    def test_method_mean(self, built_exp_inner):
        sc = StackingCollector('stk_mean', Connector(
            edges={'y': [(None, 'target')]}
        ), output_var=None, experimenter=built_exp_inner, method='mean')
        built_exp_inner.add_collector(sc)
        assert sc.has_node('dt')

    def test_reset_nodes(self, built_exp):
        sc = StackingCollector('stk', Connector(), output_var=None,
                               experimenter=built_exp)
        built_exp.add_collector(sc)
        assert sc.has_node('dt')
        sc.reset_nodes(['dt'])
        assert not sc.has_node('dt')

    def test_save_load(self, built_exp):
        sc = StackingCollector('stk', Connector(
            edges={'y': [(None, 'target')]}
        ), output_var=None, experimenter=built_exp)
        built_exp.add_collector(sc)
        loaded = StackingCollector.load(sc.path)
        assert loaded.has_node('dt')
        ds_orig = sc.get_dataset()
        ds_loaded = loaded.get_dataset()
        pd.testing.assert_frame_equal(ds_orig, ds_loaded)

    def test_index_preserved(self, built_exp):
        sc = StackingCollector('stk', Connector(
            edges={'y': [(None, 'target')]}
        ), output_var=None, experimenter=built_exp)
        built_exp.add_collector(sc)
        ds = sc.get_dataset()
        all_valid_idx = np.concatenate([
            built_exp.valid_idx_list[i]
            for i in range(built_exp.get_n_splits())
        ])
        expected_index = built_exp.data.data.index[all_valid_idx]
        pd.testing.assert_index_equal(ds.index, expected_index)


class TestModelAttrCollector:
    def test_collect_basic(self, built_exp):
        from mllabs.adapter import DecisionTreeAdapter
        mac = ModelAttrCollector('fi', Connector(processor=DecisionTreeClassifier),
                                result_key='feature_importances',
                                adapter=DecisionTreeAdapter())
        built_exp.add_collector(mac)
        assert mac.has('dt')

    def test_get_attr(self, built_exp):
        from mllabs.adapter import DecisionTreeAdapter
        mac = ModelAttrCollector('fi', Connector(processor=DecisionTreeClassifier),
                                result_key='feature_importances',
                                adapter=DecisionTreeAdapter())
        built_exp.add_collector(mac)
        result = mac.get_attr('dt')
        assert isinstance(result, list)
        assert len(result) == 2

    def test_get_attr_idx(self, built_exp):
        from mllabs.adapter import DecisionTreeAdapter
        mac = ModelAttrCollector('fi', Connector(processor=DecisionTreeClassifier),
                                result_key='feature_importances',
                                adapter=DecisionTreeAdapter())
        built_exp.add_collector(mac)
        result = mac.get_attr('dt', idx=0)
        assert isinstance(result, list)

    def test_get_attrs(self, multi_head_exp):
        from mllabs.adapter import DecisionTreeAdapter
        mac = ModelAttrCollector('fi', Connector(processor=DecisionTreeClassifier),
                                result_key='feature_importances',
                                adapter=DecisionTreeAdapter())
        multi_head_exp.add_collector(mac)
        result = mac.get_attrs()
        assert 'dt1' in result
        assert 'dt2' in result

    def test_get_attrs_agg(self, built_exp):
        from mllabs.adapter import DecisionTreeAdapter
        mac = ModelAttrCollector('fi', Connector(processor=DecisionTreeClassifier),
                                result_key='feature_importances',
                                adapter=DecisionTreeAdapter())
        built_exp.add_collector(mac)
        result = mac.get_attrs_agg('dt')
        assert isinstance(result, pd.Series)

    def test_get_attrs_agg_inner_only(self, built_exp):
        from mllabs.adapter import DecisionTreeAdapter
        mac = ModelAttrCollector('fi', Connector(processor=DecisionTreeClassifier),
                                result_key='feature_importances',
                                adapter=DecisionTreeAdapter())
        built_exp.add_collector(mac)
        result = mac.get_attrs_agg('dt', agg_inner=True, agg_outer=False)
        assert isinstance(result, pd.DataFrame)

    def test_get_attrs_agg_invalid(self, built_exp):
        from mllabs.adapter import DecisionTreeAdapter
        mac = ModelAttrCollector('fi', Connector(processor=DecisionTreeClassifier),
                                result_key='feature_importances',
                                adapter=DecisionTreeAdapter())
        built_exp.add_collector(mac)
        with pytest.raises(ValueError):
            mac.get_attrs_agg('dt', agg_inner=False, agg_outer=True)

    def test_not_mergeable(self, built_exp):
        from mllabs.adapter import DecisionTreeAdapter
        mac = ModelAttrCollector('tree', Connector(processor=DecisionTreeClassifier),
                                result_key='tree',
                                adapter=DecisionTreeAdapter())
        built_exp.add_collector(mac)
        with pytest.raises(ValueError, match='not mergeable'):
            mac.get_attrs_agg('dt')

    def test_reset_nodes(self, built_exp):
        from mllabs.adapter import DecisionTreeAdapter
        mac = ModelAttrCollector('fi', Connector(processor=DecisionTreeClassifier),
                                result_key='feature_importances',
                                adapter=DecisionTreeAdapter())
        built_exp.add_collector(mac)
        mac.reset_nodes(['dt'])
        assert not mac.has('dt')

    def test_save_load(self, built_exp):
        from mllabs.adapter import DecisionTreeAdapter
        mac = ModelAttrCollector('fi', Connector(processor=DecisionTreeClassifier),
                                result_key='feature_importances',
                                adapter=DecisionTreeAdapter())
        built_exp.add_collector(mac)
        loaded = ModelAttrCollector.load(mac.path)
        assert loaded.has('dt')

    def test_auto_adapter(self):
        mac = ModelAttrCollector('fi', Connector(processor=DecisionTreeClassifier),
                                result_key='feature_importances')
        assert mac.adapter is not None

    def test_auto_adapter_invalid_key(self):
        with pytest.raises(RuntimeError):
            ModelAttrCollector('fi', Connector(processor=DecisionTreeClassifier),
                               result_key='nonexistent_key')


class TestOutputCollector:
    def test_collect_basic(self, built_exp):
        oc = OutputCollector('out', Connector(), output_var=None)
        built_exp.add_collector(oc)
        assert oc.has_node('dt')

    def test_get_output(self, built_exp):
        oc = OutputCollector('out', Connector(), output_var=None)
        built_exp.add_collector(oc)
        result = oc.get_output('dt', 0, 0)
        assert 'output_valid' in result
        assert 'output_train' in result
        assert 'columns' in result

    def test_get_output_structure(self, built_exp):
        oc = OutputCollector('out', Connector(), output_var=None)
        built_exp.add_collector(oc)
        result = oc.get_output('dt', 0, 0)
        assert isinstance(result['output_valid'], np.ndarray)
        assert isinstance(result['output_train'], tuple)
        assert len(result['output_train']) == 2

    def test_get_outputs(self, built_exp):
        oc = OutputCollector('out', Connector(), output_var=None)
        built_exp.add_collector(oc)
        results = oc.get_outputs('dt')
        assert isinstance(results, dict)
        assert len(results) == built_exp.get_n_splits() * built_exp.get_n_splits_inner()
        for key in results:
            assert isinstance(key, tuple)
            assert len(key) == 2

    def test_get_outputs_inner_split(self, built_exp_inner):
        oc = OutputCollector('out', Connector(), output_var=None)
        built_exp_inner.add_collector(oc)
        results = oc.get_outputs('dt')
        n_expected = built_exp_inner.get_n_splits() * built_exp_inner.get_n_splits_inner()
        assert len(results) == n_expected

    def test_get_output_not_found(self, built_exp):
        oc = OutputCollector('out', Connector(), output_var=None)
        built_exp.add_collector(oc)
        with pytest.raises(FileNotFoundError):
            oc.get_output('dt', 99, 99)

    def test_get_outputs_node_not_found(self, built_exp):
        oc = OutputCollector('out', Connector(), output_var=None)
        built_exp.add_collector(oc)
        with pytest.raises(FileNotFoundError):
            oc.get_outputs('nonexistent')

    def test_reset_nodes(self, built_exp):
        oc = OutputCollector('out', Connector(), output_var=None)
        built_exp.add_collector(oc)
        assert oc.has_node('dt')
        oc.reset_nodes(['dt'])
        assert not oc.has_node('dt')

    def test_save_load(self, built_exp):
        oc = OutputCollector('out', Connector(), output_var=None)
        built_exp.add_collector(oc)
        loaded = OutputCollector.load(oc.path)
        assert loaded.has_node('dt')
        result_orig = oc.get_output('dt', 0, 0)
        result_loaded = loaded.get_output('dt', 0, 0)
        np.testing.assert_array_equal(result_orig['output_valid'],
                                      result_loaded['output_valid'])

    def test_saved_nodes(self, multi_head_exp):
        oc = OutputCollector('out', Connector(), output_var=None)
        multi_head_exp.add_collector(oc)
        saved = oc._get_saved_nodes()
        assert 'dt1' in saved
        assert 'dt2' in saved


class TestCollectorWithExperimenter:
    def test_collect_skip_existing(self, built_exp):
        mc = MetricCollector('acc', Connector(), output_var=None,
                             metric_func=accuracy_metric)
        built_exp.add_collector(mc)
        metric_before = mc.get_metric('dt').copy()
        built_exp.collect(mc, exist='skip')
        metric_after = mc.get_metric('dt')
        pd.testing.assert_series_equal(metric_before, metric_after)

    def test_experimenter_save_load_with_collectors(self, built_exp, sample_data):
        mc = MetricCollector('acc', Connector(), output_var=None,
                             metric_func=accuracy_metric)
        built_exp.add_collector(mc)
        path = built_exp.path

        loaded = Experimenter.load(path, sample_data)
        assert 'acc' in loaded.collectors
        loaded_mc = loaded.collectors['acc']
        assert loaded_mc.has('dt')
        result_orig = mc.get_metric('dt')
        result_loaded = loaded_mc.get_metric('dt')
        pd.testing.assert_series_equal(result_orig, result_loaded)

    def test_reset_nodes_clears_collectors(self, built_exp):
        mc = MetricCollector('acc', Connector(), output_var=None,
                             metric_func=accuracy_metric)
        built_exp.add_collector(mc)
        assert mc.has('dt')
        built_exp.reset_nodes(['dt'])
        assert not mc.has('dt')

    def test_multiple_collectors(self, built_exp):
        from mllabs.adapter import DecisionTreeAdapter
        mc = MetricCollector('acc', Connector(), output_var=None,
                             metric_func=accuracy_metric)
        oc = OutputCollector('out', Connector(), output_var=None)
        mac = ModelAttrCollector('fi', Connector(processor=DecisionTreeClassifier),
                                result_key='feature_importances',
                                adapter=DecisionTreeAdapter())
        built_exp.add_collector(mc)
        built_exp.add_collector(oc)
        built_exp.add_collector(mac)
        assert mc.has('dt')
        assert oc.has_node('dt')
        assert mac.has('dt')


class TestBaseCollector:
    def test_get_nodes_none(self):
        from mllabs.collector._base import Collector
        c = Collector('test', Connector())
        result = c._get_nodes(None, ['a', 'b', 'c'])
        assert result == ['a', 'b', 'c']

    def test_get_nodes_list(self):
        from mllabs.collector._base import Collector
        c = Collector('test', Connector())
        result = c._get_nodes(['a', 'c'], ['a', 'b', 'c'])
        assert result == ['a', 'c']

    def test_get_nodes_list_filter(self):
        from mllabs.collector._base import Collector
        c = Collector('test', Connector())
        result = c._get_nodes(['a', 'x'], ['a', 'b', 'c'])
        assert result == ['a']

    def test_get_nodes_regex(self):
        from mllabs.collector._base import Collector
        c = Collector('test', Connector())
        result = c._get_nodes('dt', ['dt1', 'dt2', 'scaler'])
        assert result == ['dt1', 'dt2']

    def test_get_nodes_invalid_type(self):
        from mllabs.collector._base import Collector
        c = Collector('test', Connector())
        with pytest.raises(ValueError):
            c._get_nodes(123, ['a', 'b'])



"""
Column selection helper functions for resolve_columns
"""
import re

def ohe_drop_first(columns, processor):
    # 각 원래 변수에 대해 첫 번째 컬럼을 만났는지 추적
    org_X = processor.X_
    first_seen = {var: False for var in org_X}

    mask = []
    for col in columns:
        # '__' 뒤의 부분 추출
        if '__' not in col:
            mask.append(False)
            continue

        suffix = col.split('__', 1)[1]

        # 어떤 org_X 변수에 속하는지 확인
        matched = False
        for org_var in org_X:
            # suffix가 org_var로 시작하는지 확인 (예: color_red는 color로 시작)
            if suffix.startswith(f"{org_var}_"):
                matched = True
                if not first_seen[org_var]:
                    # 첫 번째 컬럼은 제외 (False)
                    mask.append(False)
                    first_seen[org_var] = True
                else:
                    # 나머지는 포함 (True)
                    mask.append(True)
                break

        # org_X의 어떤 변수에도 속하지 않으면 제외
        if not matched:
            mask.append(False)

    return mask

def _polynomial_feature_names(input_features, degree=2, interaction_only=False, include_bias=True):
    from itertools import combinations, combinations_with_replacement
    n_features = len(input_features)
    feature_names = []
    if include_bias:
        feature_names.append("1")
    comb_func = combinations if interaction_only else combinations_with_replacement
    for d in range(1, degree + 1):
        for comb in comb_func(range(n_features), d):
            counts = {}
            for idx in comb:
                counts[idx] = counts.get(idx, 0) + 1
            terms = []
            for idx, power in counts.items():
                name = input_features[idx]
                if power > 1:
                    name = f"{name}^{power}"
                terms.append(name)
            feature_names.append(" ".join(terms))
    return feature_names

def subset_poly(columns, vars, processor=None):
    obj = processor.obj
    degree = obj.degree if hasattr(obj, 'degree') else 2
    interaction_only = obj.interaction_only if hasattr(obj, 'interaction_only') else False
    include_bias = obj.include_bias if hasattr(obj, 'include_bias') else True

    org_X = processor.X_
    vars_ = list()
    for x in org_X:
        for i in vars:
            if i == x or re.match(i, x):
                vars_.append(x)
                break

    subset_names = set(_polynomial_feature_names(
        vars_, degree=degree, interaction_only=interaction_only, include_bias=include_bias
    ))

    node_name = processor.name
    prefix = f"{node_name}__"

    mask = []
    for col in columns:
        if col.startswith(prefix):
            suffix = col[len(prefix):]
            mask.append(suffix in subset_names)
        else:
            mask.append(False)
    return mask

def get_origin_var(columns, org_X):
    l = list()
    for col in columns:
        suffix = col.split('__', 1)[-1]
        org = "Unknown"
        for org_var in org_X:
            if suffix.startswith(f'{org_var}_') or suffix == org_var:
                org = org_var
                break
        l.append(org)
    return l



import pickle as pkl
from pathlib import Path

from ._data_wrapper import wrap
from ._node_processor import resolve_columns


class Inferencer:

    def __init__(self, pipeline, selected_stages, selected_heads, n_splits, node_objs, v=None):
        self.pipeline = pipeline
        self.selected_stages = selected_stages
        self.selected_heads = selected_heads
        self.n_splits = n_splits
        self.node_objs = node_objs
        self.v = v

    def process(self, data, agg='mean'):
        results = list(self._process_splits(data))

        if self.n_splits == 1:
            return results[0]

        if agg is None:
            return results
        elif agg == 'mean':
            return type(results[0]).mean(iter(results))
        elif agg == 'mode':
            return type(results[0]).mode(iter(results))
        elif callable(agg):
            return agg(results)
        else:
            raise ValueError(f"Unknown agg: {agg}")

    def _process_splits(self, data):
        data = wrap(data)
        ordered = [
            name for name in self.pipeline._get_affected_nodes([None])
            if name in set(self.selected_stages + self.selected_heads)
        ]
        stage_set = set(self.selected_stages)

        for split_idx in range(self.n_splits):
            data_dicts = {None: (data, None)}
            head_outputs = []

            for name in ordered:
                obj = self.node_objs[name][split_idx]
                node_attrs = self.pipeline.get_node_attrs(name)
                output = self._process_node(obj, data_dicts, node_attrs['edges'])

                if name in stage_set:
                    data_dicts[name] = (output, obj)
                else:
                    if self.v is not None:
                        cols = resolve_columns(output, self.v, processor=obj)
                        output = output.select_columns(cols)
                    head_outputs.append(output)

            if len(head_outputs) == 1:
                yield head_outputs[0]
            else:
                yield type(head_outputs[0]).concat(head_outputs, axis=1)

    def _process_node(self, obj, data_dicts, edges):
        input_data = self._get_process_data(data_dicts, edges)
        return obj.process(input_data['X'])

    def _get_process_data(self, data_dicts, edges):
        result = {}
        for key, edge_list in edges.items():
            parts = []
            for src_node, var in edge_list:
                src, obj = data_dicts[src_node]
                if var is not None:
                    cols = resolve_columns(src, var, processor=obj)
                    src = src.select_columns(cols)
                parts.append(src)
            if len(parts) == 1:
                result[key] = parts[0]
            else:
                result[key] = type(parts[0]).concat(parts, axis=1)
        return result

    # ------------------------------------------------------------------
    # save / load
    # ------------------------------------------------------------------

    def save(self, path):
        path = Path(path)
        path.mkdir(parents=True, exist_ok=True)
        save_data = {
            'pipeline': self.pipeline,
            'selected_stages': self.selected_stages,
            'selected_heads': self.selected_heads,
            'n_splits': self.n_splits,
            'node_objs': self.node_objs,
            'v': self.v,
        }
        with open(path / '__inferencer.pkl', 'wb') as f:
            pkl.dump(save_data, f)

    @classmethod
    def load(cls, path):
        path = Path(path)
        with open(path / '__inferencer.pkl', 'rb') as f:
            save_data = pkl.load(f)
        return cls(
            save_data['pipeline'],
            save_data['selected_stages'],
            save_data['selected_heads'],
            save_data['n_splits'],
            save_data['node_objs'],
            save_data.get('v'),
        )



import pickle
import pandas as pd

from ._base import Collector
from .._node_processor import resolve_columns


class MetricCollector(Collector):
    def __init__(self, name, connector, output_var, metric_func, include_train=False):
        super().__init__(name, connector)
        self.output_var = output_var
        self.metric_func = metric_func
        self.include_train = include_train
        self.metrics = {}
        self._sub = {}

    def _start(self, node):
        self.metrics[node] = []
        self._sub[node] = []

    def _collect(self, node, idx, inner_idx, context):
        if 'y' not in context['input']:
            self._sub[node].append(None)
            return

        (true_t, true_tv), true_v = context['input']['y']
        cols = resolve_columns(context['output_valid'], self.output_var)
        if len(cols) == 0:
            self._sub[node].append(None)
            return

        prd_v = context['output_valid'].select_columns(cols)
        result = {'valid': self.metric_func(true_v.data, prd_v.data)}

        if self.include_train:
            prd_t = context['output_train'][0].select_columns(cols)
            result['train_sub'] = self.metric_func(true_t.data, prd_t.data)
            if true_tv is not None:
                prd_tv = context['output_train'][1].select_columns(cols)
                result['valid_sub'] = self.metric_func(true_tv.data, prd_tv.data)

        self._sub[node].append(result)

    def _end_idx(self, node, idx):
        self.metrics[node].append(self._sub[node])
        self._sub[node] = []

    def _end(self, node):
        if len(self.metrics[node]) == 0:
            del self.metrics[node]
        if node in self._sub:
            del self._sub[node]
        self.save()

    def has_node(self, node):
        return node in self.metrics

    def reset_nodes(self, nodes):
        for node in nodes:
            if node in self.metrics:
                del self.metrics[node]
        self.save()

    def save(self):
        if self.path is None:
            return
        self._ensure_path()
        data = {
            'name': self.name,
            'connector': self.connector,
            'output_var': self.output_var,
            'metric_func': self.metric_func,
            'include_train': self.include_train,
            'metrics': self.metrics,
        }
        with open(self.path / '__config.pkl', 'wb') as f:
            pickle.dump(data, f)

    @classmethod
    def load(cls, path):
        with open(path / '__config.pkl', 'rb') as f:
            data = pickle.load(f)
        obj = cls(
            name=data['name'],
            connector=data['connector'],
            output_var=data['output_var'],
            metric_func=data['metric_func'],
            include_train=data['include_train'],
        )
        obj.metrics = data['metrics']
        obj.path = path
        return obj

    def get_metric(self, node):
        l = list()
        for i, sub in enumerate(self.metrics[node]):
            l.append(
                pd.concat([pd.Series(j, name=str(no)) for no, j in enumerate(sub)], axis=1).unstack()
            )
        return pd.concat(l, axis=1).unstack(level=[0, 1]).rename(node)

    def get_metrics(self, nodes=None):
        node_names = self._get_nodes(nodes, self.metrics.keys())
        return pd.concat([self.get_metric(node) for node in node_names], axis=1).T

    def get_metrics_agg(self, nodes=None, inner_fold=True, outer_fold=True, include_std=False):
        if outer_fold and not inner_fold:
            raise ValueError("")
        df = self.get_metrics(nodes)
        if inner_fold:
            df_agg_mean = df.stack(level=1, future_stack=True).groupby(level=0).mean()
            if include_std:
                df_agg_std = df.stack(level=1, future_stack=True).groupby(level=0).std()
            else:
                df_agg_std = None
            if outer_fold:
                df_agg_mean = df_agg_mean.stack(level=0, future_stack=True).groupby(level=0).mean()
                if include_std:
                    df_agg_std = df_agg_std.stack(level=0, future_stack=True).groupby(level=0).mean()
            return df_agg_mean, df_agg_std
        return df



__version__ = "0.3.0"

from ._experimenter import Experimenter
from ._inferencer import Inferencer
from ._connector import Connector
from .collector import Collector, MetricCollector, StackingCollector, ModelAttrCollector, SHAPCollector, OutputCollector
from .filter import DataFilter, RandomFilter, IndexFilter

__all__ = [
    'Experimenter',
    'Inferencer',
    'Connector',
    'Collector',
    'MetricCollector',
    'StackingCollector',
    'ModelAttrCollector',
    'SHAPCollector',
    'OutputCollector',
    'DataFilter',
    'RandomFilter',
    'IndexFilter',
]


"""
LightGBM adapter
"""

import pandas as pd
from ._base import ModelAdapter
from lightgbm import early_stopping

def create_progress_callback(n_estimators, period_pct, logger):
    last_printed = [-1]  # mutable object
    def callback(env):
        current = env.iteration + 1
        percentage = (current / n_estimators) * 100

        # period_pct마다 출력
        if int(percentage / (period_pct * 100)) > last_printed[0]:
            last_printed[0] = int(percentage / (period_pct * 100))

            # metric 정보 추출
            metrics_str = ""
            if env.evaluation_result_list:
                last_metrics = []
                for item in env.evaluation_result_list:
                    dataset_name, metric_name, value, _ = item
                    last_metrics.append(f"{dataset_name}-{metric_name}: {value:.4f}")
                metrics_str = ", ".join(last_metrics)
            logger.adhoc_progress(current, n_estimators, metrics_str if metrics_str else None)

    return callback

class LightGBMAdapter(ModelAdapter):
    """Adapter for LightGBM models (LGBMClassifier, LGBMRegressor)

    LightGBM도 eval_set 파라미터를 사용하지만 약간 다른 방식입니다.
    """
    def get_params(self, params, logger = None):
        """모델 생성자에 전달할 파라미터를 조정

        Args:
            params (dict): 원본 파라미터

        Returns:
            dict: 조정된 파라미터
        """
        return {k: v for k, v in params.items() if k not in ['early_stopping', 'eval_metric']}

    def get_fit_params(self, data_dict, X, y=None, params=None, logger=None):
        """LightGBM의 fit 파라미터 구성"""
        from .._data_wrapper import unwrap

        fit_params = {}

        # data_dict에서 데이터 추출
        train_X, train_v_X = data_dict['X']
        if 'y' in data_dict:
            train_y, train_v_y = data_dict['y']
        else:
            train_y, train_v_y = None, None

        # eval_set 구성
        if self.eval_mode and self.eval_mode != 'none' and train_v_X is not None and train_v_y is not None:
            train_X_native = unwrap(train_X)
            train_y_native = unwrap(train_y)
            train_v_X_native = unwrap(train_v_X)
            train_v_y_native = unwrap(train_v_y)

            if self.eval_mode == 'valid':
                fit_params['eval_set'] = [(train_v_X_native, train_v_y_native)]
            elif self.eval_mode == 'both':
                fit_params['eval_set'] = [(train_X_native, train_y_native), (train_v_X_native, train_v_y_native)]

        # verbose 처리
        if self.verbose > 0:
            if self.verbose < 1:
                # 0 < verbose < 1: 진행률 기반 출력
                # n_estimators 추출 (기본값 100)
                n_estimators = params.get('n_estimators', 100) if params else 100
                callbacks = fit_params.get('callbacks', [])
                if logger is not None:
                    callbacks.append(create_progress_callback(n_estimators, self.verbose, logger))
                fit_params['callbacks'] = callbacks
            else:
                # verbose >= 1: LightGBM 기본 verbose (iteration 단위)
                fit_params['verbose'] = int(self.verbose)
        
        if 'early_stopping' in params:
            if 'callbacks' not in fit_params:
                fit_params['callbacks'] = list()
            fit_params['callbacks'].append(params['early_stopping'])
        if 'eval_metric' in params:
            fit_params['eval_metric'] = params['eval_metric']
        return fit_params

    @staticmethod
    def _get_feature_importances(processor):
        obj = processor.obj
        input_vars = list(processor.X_) if hasattr(processor, 'X_') and processor.X_ is not None else list(range(obj.n_features_in_))

        return pd.Series(
            obj.feature_importances_,
            index=input_vars, name = 'importance'
        )

    @staticmethod
    def _get_evals_result(processor):
        obj = processor.obj
        evals_result = obj.evals_result_ if hasattr(obj, 'evals_result_') else {}
        return pd.concat(
            [pd.DataFrame(v).stack().rename(k) for k, v in evals_result.items()], axis=1
        ).stack()

    @staticmethod
    def _get_trees(processor):
        obj = processor.obj
        dump = obj.booster_.dump_model()
        return dump.get('tree_info', [])

LightGBMAdapter.result_objs = {
    'feature_importances': (LightGBMAdapter._get_feature_importances, True),
    'evals_result': (LightGBMAdapter._get_evals_result, True),
    'trees': (LightGBMAdapter._get_trees, False)
}


"""
Data Wrapper for supporting multiple data libraries (pandas, Polars, cudf, numpy)

Provides a unified interface for common DataFrame operations across different libraries.
"""

from abc import ABC, abstractmethod
import numpy as np

try:
    import pandas as pd
except Exception:
    pd = None

try:
    import polars as pl
except Exception:
    pl = None


class DataWrapper(ABC):
    """Abstract base class for data wrappers

    Provides unified interface for common operations across different data libraries.
    """

    def __init__(self, data):
        """Initialize wrapper with native data object"""
        self.data = data

    @abstractmethod
    def iloc(self, indices):
        """Select rows by integer indices

        Args:
            indices: Integer indices (list, array, or slice)

        Returns:
            DataWrapper: New wrapper with selected rows
        """
        pass

    @abstractmethod
    def select_columns(self, columns):
        """Select specific columns

        Args:
            columns: Column names (list or single column name)

        Returns:
            DataWrapper: New wrapper with selected columns
        """
        pass

    @abstractmethod
    def get_columns(self):
        """Get list of column names

        Returns:
            list: List of column names
        """
        pass

    @abstractmethod
    def get_shape(self):
        """Get shape of data

        Returns:
            tuple: (n_rows, n_cols)
        """
        pass

    @abstractmethod
    def get_index(self):
        """Get index of data

        Returns:
            Index object (type depends on library)
        """
        pass

    @staticmethod
    @abstractmethod
    def concat(wrappers, axis=1):
        """Concatenate multiple wrappers

        Args:
            wrappers: List of DataWrapper objects
            axis: Axis to concatenate (0=rows, 1=columns)

        Returns:
            DataWrapper: Concatenated result
        """
        pass

    @abstractmethod
    def squeeze(self):
        pass

    def to_native(self):
        """Get native data object

        Returns:
            Native data object (DataFrame, ndarray, etc.)
        """
        return self.data

    @abstractmethod
    def to_array(self, array_type='ndarray'):
        pass

    @staticmethod
    def from_native(data):
        """Create appropriate wrapper from native data object

        Args:
            data: Native data object

        Returns:
            DataWrapper: Appropriate wrapper instance
        """
        if data is None:
            return None

        # Check type and return appropriate wrapper
        type_name = type(data).__name__
        module_name = type(data).__module__

        if 'pandas' in module_name:
            return PandasWrapper(data)
        elif 'polars' in module_name:
            return PolarsWrapper(data)
        elif 'cudf' in module_name:
            return CudfWrapper(data)
        elif isinstance(data, np.ndarray):
            return NumpyWrapper(data)
        else:
            # Default to pandas if unknown
            import pandas as pd
            if isinstance(data, pd.DataFrame) or isinstance(data, pd.Series):
                return PandasWrapper(data)
            raise TypeError(f"Unsupported data type: {type(data)}")

    @classmethod
    @abstractmethod
    def from_output(cls, output, column_names=None, index=None):
        """Create wrapper instance from processor output with optional column names and index

        Processor의 출력(numpy array, DataFrame 등)을 받아서 해당 wrapper의
        native 타입으로 변환하고 컬럼명과 인덱스를 설정한 후 wrapper 인스턴스를 반환합니다.

        Args:
            output: Processor의 출력 (numpy array, DataFrame, etc.)
            column_names: 설정할 컬럼명 리스트 (None이면 자동 생성 또는 유지)
            index: 설정할 인덱스 (None이면 기본 인덱스 또는 유지)

        Returns:
            DataWrapper: Wrapper 인스턴스
        """
        pass

    def __getitem__(self, key):
        """Support indexing syntax: wrapper[key]"""
        if isinstance(key, (list, str)):
            return self.select_columns(key)
        else:
            # For other types, delegate to native object
            return DataWrapper.from_native(self.data[key])

    @staticmethod
    def simple(iterator):
        """첫 번째 inner fold 결과만 반환"""
        return next(iterator)

    @staticmethod
    def mean(iterator):
        """평균값으로 집계"""
        # DataWrapper에서 native 추출
        ret = unwrap(next(iterator)).copy()
        cnt = 1
        for i in iterator:
            ret += unwrap(i)
            cnt += 1
        return wrap(ret / cnt)
    
    @staticmethod
    @abstractmethod
    def mode(iterator):
        """최빈값으로 집계"""
        pass


class PandasWrapper(DataWrapper):
    """Wrapper for pandas DataFrame/Series"""

    def iloc(self, indices):
        return PandasWrapper(self.data.iloc[indices])

    def select_columns(self, columns):
        return PandasWrapper(self.data[columns])

    def get_columns(self):
        if hasattr(self.data, 'columns'):
            return self.data.columns.tolist()
        else:
            # Series has name
            return [self.data.name] if self.data.name is not None else [0]

    def get_shape(self):
        return self.data.shape

    def get_index(self):
        return self.data.index

    @staticmethod
    def concat(wrappers, axis=1):
        import pandas as pd
        native_list = [w.to_native() for w in wrappers]
        result = pd.concat(native_list, axis=axis)
        return PandasWrapper(result)

    @classmethod
    def from_output(cls, output, column_names=None, index=None):
        """Convert processor output to PandasWrapper

        Args:
            output: Processor 출력 (numpy array, DataFrame, etc.)
            column_names: 컬럼명 리스트
            index: 인덱스

        Returns:
            PandasWrapper 인스턴스
        """

        if output is None:
            return None

        # 이미 pandas DataFrame인 경우
        if isinstance(output, pd.DataFrame):
            if column_names is not None:
                output.columns = column_names
            if index is not None:
                output.index = index
            return cls(output)

        # numpy array인 경우 → DataFrame으로 변환
        elif isinstance(output, np.ndarray):
            if output.ndim == 1:
                if column_names is None:
                    column_names = [0]
                df = pd.DataFrame({column_names[0]: output}, index=index)
            else:
                if column_names is None:
                    column_names = list(range(output.shape[1]))
                df = pd.DataFrame(output, columns=column_names, index=index)
            return cls(df)

        # 기타 DataFrame (polars, cudf) → pandas로 변환
        elif hasattr(output, 'to_pandas'):
            df = output.to_pandas()
            if column_names is not None:
                df.columns = column_names
            if index is not None:
                df.index = index
            return cls(df)

        # 변환 불가능한 경우 에러 발생
        else:
            raise TypeError(f"Cannot convert {type(output)} to pandas DataFrame")
    
    def squeeze(self):
        return PandasWrapper(self.data.squeeze())

    @staticmethod
    def mode(iterator):
        data_list = [unwrap(i) for i in iterator]
        if len(data_list) == 0:
            return None
        if len(data_list) == 1:
            return data_list[0]

        # DataFrame인 경우
        if isinstance(data_list[0], pd.DataFrame):
            result = list()
            for i in data_list[0].columns:
                result.append(
                    pd.concat([j[i] for j in data_list], axis = 1).mode(axis=1)[0].rename(i)
                )
            return wrap(pd.concat(result, axis=1))

        # Series인 경우
        elif isinstance(data_list[0], pd.Series):
            return wrap(pd.concat(data_list).mode(axis = 1)[0])

    def to_array(self, array_type='ndarray'):
        if array_type == 'ndarray':
            return self.data.to_numpy()
        raise ValueError(f"Unsupported array_type: {array_type}")

class PolarsWrapper(DataWrapper):
    """Wrapper for Polars DataFrame"""

    def iloc(self, indices):
        import polars as pl
        if isinstance(indices, slice):
            result = self.data[indices]
        elif isinstance(indices, (list, np.ndarray)):
            result = self.data[indices]
        else:
            result = self.data[indices]
        return PolarsWrapper(result)

    def select_columns(self, columns):
        return PolarsWrapper(self.data[columns])

    def get_columns(self):
        return self.data.columns

    def get_shape(self):
        return self.data.shape

    def get_index(self):
        # Polars doesn't have index, return range
        return range(len(self.data))

    @staticmethod
    def concat(wrappers, axis=1):
        import polars as pl
        native_list = [w.to_native() for w in wrappers]
        if axis == 1:
            result = pl.concat(native_list, how="horizontal")
        else:
            result = pl.concat(native_list, how="vertical")
        return PolarsWrapper(result)

    @classmethod
    def from_output(cls, output, column_names=None, index=None):
        """Convert processor output to PolarsWrapper

        Args:
            output: Processor 출력 (numpy array, DataFrame, etc.)
            column_names: 컬럼명 리스트
            index: 인덱스 (polars는 인덱스 개념이 없으므로 무시됨)

        Returns:
            PolarsWrapper 인스턴스
        """
        import polars as pl

        if output is None:
            return None

        # 이미 polars DataFrame인 경우
        if isinstance(output, pl.DataFrame):
            if column_names is not None:
                output.columns = column_names
            return cls(output)

        # numpy array인 경우 → polars DataFrame으로 변환
        elif isinstance(output, np.ndarray):
            if output.ndim == 1:
                if column_names is None:
                    column_names = [0]
                df = pl.DataFrame({str(column_names[0]): output})
            else:
                if column_names is None:
                    column_names = [str(i) for i in range(output.shape[1])]
                else:
                    column_names = [str(c) for c in column_names]
                df = pl.DataFrame(output, schema=column_names)
            return cls(df)

        # pandas DataFrame → polars로 변환
        elif hasattr(output, 'to_numpy') and hasattr(output, 'columns'):
            df = pl.from_pandas(output)
            if column_names is not None:
                df.columns = column_names
            return cls(df)

        # 변환 불가능한 경우 에러 발생
        else:
            raise TypeError(f"Cannot convert {type(output)} to polars DataFrame")

    @staticmethod
    def mean(iterator):
        """평균값으로 집계"""
        # DataWrapper에서 native 추출
        ret = unwrap(next(iterator)).clone()
        cnt = 1
        for i in iterator:
            ret += unwrap(i)
            cnt += 1
        return wrap(ret / cnt)

    def squeeze(self):
        if isinstance(self.data, pl.DataFrame) and self.data.shape[1] == 1:
            return PolarsWrapper(self.data.to_series())
        return self

    @staticmethod
    def mode(iterator):
        data_list = [unwrap(i) for i in iterator]
        if len(data_list) == 0:
            return None
        if len(data_list) == 1:
            return wrap(data_list[0])

        # DataFrame인 경우
        if isinstance(data_list[0], pl.DataFrame):
            result = list()
            for col_name in data_list[0].columns:
                # 각 DataFrame에서 해당 컬럼을 추출하여 수평으로 결합
                combined = pl.concat([df.select(col_name).rename({col_name: f"col_{idx}"})
                                     for idx, df in enumerate(data_list)], how="horizontal")
                # 각 행별로 최빈값 계산 (polars는 행별 mode가 없으므로 pandas 변환 후 계산)
                mode_values = combined.to_pandas().mode(axis=1)[0].values
                result.append(pl.Series(col_name, mode_values))
            return wrap(pl.DataFrame(result))

        # Series인 경우
        elif isinstance(data_list[0], pl.Series):
            combined = pl.concat([s.to_frame().rename({s.name: f"col_{idx}"})
                                 for idx, s in enumerate(data_list)], how="horizontal")
            mode_values = combined.to_pandas().mode(axis=1)[0].values
            return wrap(pl.Series(data_list[0].name, mode_values))

    def to_array(self, array_type='ndarray'):
        if array_type == 'ndarray':
            return self.data.to_numpy()
        raise ValueError(f"Unsupported array_type: {array_type}")

class CudfWrapper(DataWrapper):
    """Wrapper for cuDF DataFrame (GPU-accelerated)"""

    def iloc(self, indices):
        return CudfWrapper(self.data.iloc[indices])

    def select_columns(self, columns):
        return CudfWrapper(self.data[columns])

    def get_columns(self):
        if hasattr(self.data, 'columns'):
            return self.data.columns.tolist()
        else:
            return [self.data.name] if self.data.name is not None else [0]

    def get_shape(self):
        return self.data.shape

    def get_index(self):
        return self.data.index

    @staticmethod
    def concat(wrappers, axis=1):
        import cudf
        native_list = [w.to_native() for w in wrappers]
        result = cudf.concat(native_list, axis=axis)
        return CudfWrapper(result)

    @classmethod
    def from_output(cls, output, column_names=None, index=None):
        """Convert processor output to CudfWrapper

        Args:
            output: Processor 출력 (numpy array, DataFrame, etc.)
            column_names: 컬럼명 리스트
            index: 인덱스

        Returns:
            CudfWrapper 인스턴스
        """
        import cudf

        if output is None:
            return None

        # 이미 cudf DataFrame인 경우
        if isinstance(output, cudf.DataFrame):
            if column_names is not None:
                output.columns = column_names
            if index is not None:
                output.index = index
            return cls(output)

        # numpy array인 경우 → cudf DataFrame으로 변환
        elif isinstance(output, np.ndarray):
            if output.ndim == 1:
                if column_names is None:
                    column_names = [0]
                df = cudf.DataFrame({column_names[0]: output}, index=index)
            else:
                if column_names is None:
                    column_names = list(range(output.shape[1]))
                df = cudf.DataFrame(output, columns=column_names, index=index)
            return cls(df)

        # pandas DataFrame → cudf로 변환
        elif hasattr(output, 'to_numpy') and hasattr(output, 'columns'):
            df = cudf.from_pandas(output)
            if column_names is not None:
                df.columns = column_names
            if index is not None:
                df.index = index
            return cls(df)

        # 변환 불가능한 경우 에러 발생
        else:
            raise TypeError(f"Cannot convert {type(output)} to cudf DataFrame")

    def squeeze(self):
        return CudfWrapper(self.data.squeeze())

    @staticmethod
    def mode(iterator):
        import cudf
        data_list = [unwrap(i) for i in iterator]
        if len(data_list) == 0:
            return None
        if len(data_list) == 1:
            return wrap(data_list[0])

        # DataFrame인 경우
        if isinstance(data_list[0], cudf.DataFrame):
            result = list()
            for col_name in data_list[0].columns:
                # cudf는 pandas와 유사한 API를 가지므로 pandas로 변환하여 처리
                combined = pd.concat([j[col_name].to_pandas() for j in data_list], axis=1)
                mode_series = combined.mode(axis=1)[0].rename(col_name)
                result.append(cudf.Series(mode_series))
            return wrap(cudf.concat(result, axis=1))

        # Series인 경우
        elif isinstance(data_list[0], cudf.Series):
            combined = pd.concat([s.to_pandas() for s in data_list], axis=1)
            return wrap(cudf.Series(combined.mode(axis=1)[0]))

    def to_array(self, array_type='ndarray'):
        if array_type == 'ndarray':
            return self.data.to_numpy()
        raise ValueError(f"Unsupported array_type: {array_type}")

class NumpyWrapper(DataWrapper):
    """Wrapper for NumPy ndarray"""

    def __init__(self, data, columns=None):
        """Initialize with ndarray and optional column names

        Args:
            data: numpy ndarray
            columns: Optional list of column names
        """
        super().__init__(data)
        if columns is None:
            # Generate default column names
            if data.ndim == 1:
                self.columns = [0]
            else:
                self.columns = list(range(data.shape[1]))
        else:
            self.columns = columns

    def iloc(self, indices):
        if isinstance(indices, slice):
            result = self.data[indices]
        elif isinstance(indices, (list, np.ndarray)):
            result = self.data[indices]
        else:
            result = self.data[indices]
        return NumpyWrapper(result, self.columns)

    def select_columns(self, columns):
        if isinstance(columns, str):
            # Single column name
            if columns in self.columns:
                col_idx = self.columns.index(columns)
                if self.data.ndim == 1:
                    return NumpyWrapper(self.data, [columns])
                else:
                    return NumpyWrapper(self.data[:, col_idx], [columns])
            else:
                raise KeyError(f"Column '{columns}' not found")
        else:
            # Multiple columns
            col_indices = [self.columns.index(c) for c in columns]
            if self.data.ndim == 1:
                return NumpyWrapper(self.data, columns)
            else:
                return NumpyWrapper(self.data[:, col_indices], columns)

    def get_columns(self):
        return self.columns

    def get_shape(self):
        return self.data.shape

    def get_index(self):
        # NumPy doesn't have index, return range
        return range(len(self.data))

    @staticmethod
    def concat(wrappers, axis=1):
        native_list = [w.to_native() for w in wrappers]
        result = np.concatenate(native_list, axis=axis)

        # Merge column names
        if axis == 1:
            all_columns = []
            for w in wrappers:
                all_columns.extend(w.columns)
            return NumpyWrapper(result, all_columns)
        else:
            # Use first wrapper's columns for vertical concat
            return NumpyWrapper(result, wrappers[0].columns)

    @classmethod
    def from_output(cls, output, column_names=None, index=None):
        """Convert processor output to NumpyWrapper

        Args:
            output: Processor 출력 (numpy array, DataFrame, etc.)
            column_names: 컬럼명 리스트
            index: 인덱스 (numpy는 인덱스 개념이 없으므로 무시됨)

        Returns:
            NumpyWrapper 인스턴스
        """
        if output is None:
            return None

        # 이미 numpy array인 경우
        if isinstance(output, np.ndarray):
            if column_names is None:
                if output.ndim == 1:
                    column_names = [0]
                else:
                    column_names = list(range(output.shape[1]))
            return cls(output, column_names)

        # DataFrame인 경우 → numpy array로 변환
        elif hasattr(output, 'to_numpy'):
            arr = output.to_numpy()
            if column_names is None:
                if hasattr(output, 'columns'):
                    column_names = output.columns.tolist()
                else:
                    column_names = list(range(arr.shape[1] if arr.ndim > 1 else 1))
            return cls(arr, column_names)

        # 변환 불가능한 경우 에러 발생
        else:
            raise TypeError(f"Cannot convert {type(output)} to numpy array")
        
    def squeeze(self):
        return NumpyWrapper(np.squeeze(self.data))

    @staticmethod
    def mode(iterator):
        from scipy import stats
        data_list = [unwrap(i) for i in iterator]
        if len(data_list) == 0:
            return None
        if len(data_list) == 1:
            return wrap(data_list[0])

        # 모든 array를 stack하여 (n_samples, n_arrays) 또는 (n_samples, n_cols, n_arrays) 형태로 만듦
        stacked = np.stack(data_list, axis=-1)

        # scipy.stats.mode를 사용하여 마지막 축(n_arrays)에서 최빈값 계산
        mode_result = stats.mode(stacked, axis=-1, keepdims=False)
        return wrap(mode_result.mode)

    def to_array(self, array_type='ndarray'):
        if array_type == 'ndarray':
            return self.data
        raise ValueError(f"Unsupported array_type: {array_type}")


def wrap(data):
    """Convenience function to wrap native data

    Args:
        data: Native data object

    Returns:
        DataWrapper: Appropriate wrapper instance
    """
    return DataWrapper.from_native(data)


def unwrap(wrapper):
    """Convenience function to unwrap to native data

    Args:
        wrapper: DataWrapper instance

    Returns:
        Native data object
    """
    if wrapper is None:
        return None
    if isinstance(wrapper, DataWrapper):
        return wrapper.to_native()
    else:
        # Already native
        return wrapper



import re


class Collector:
    def __init__(self, name, connector):
        self.name = name
        self.connector = connector
        self.path = None

    def _start(self, node):
        pass

    def _collect(self, node, idx, inner_idx, context):
        pass

    def _end_idx(self, node, idx):
        pass

    def _end(self, node):
        pass

    def has(self, node):
        return self.has_node(node)

    def has_node(self, node):
        return False

    def reset_nodes(self, nodes):
        pass

    def _ensure_path(self):
        if self.path is not None and not self.path.exists():
            self.path.mkdir(parents=True, exist_ok=True)

    def save(self):
        pass

    @classmethod
    def load(cls, path):
        raise NotImplementedError

    def _get_nodes(self, nodes, available_nodes):
        if nodes is None:
            return list(available_nodes)
        elif isinstance(nodes, list):
            return [n for n in nodes if n in available_nodes]
        elif isinstance(nodes, str):
            pat = re.compile(nodes)
            return [k for k in available_nodes if k is not None and pat.search(k)]
        else:
            raise ValueError(f"nodes must be None, list, or str, got {type(nodes)}")



import pytest
import numpy as np
import pandas as pd
from pathlib import Path

from sklearn.preprocessing import StandardScaler
from sklearn.tree import DecisionTreeClassifier
from sklearn.model_selection import ShuffleSplit, KFold

from mllabs._experimenter import Experimenter, DataCache
from mllabs._expobj import StageObj, HeadObj
from mllabs._pipeline import Pipeline
from mllabs import Connector, MetricCollector


class BadProcessor:
    __name__ = 'BadProcessor'
    def __init__(self, **kwargs):
        pass
    def fit(self, X, y=None):
        raise ValueError("intentional error")
    def transform(self, X):
        pass


class ErrorProcessor:
    __name__ = 'ErrorProcessor'
    def __init__(self, **kwargs):
        pass
    def fit(self, X, y=None):
        raise TypeError("test error msg")
    def transform(self, X):
        pass


class BadPredictor:
    __name__ = 'BadPredictor'
    def __init__(self, **kwargs):
        pass
    def fit(self, X, y=None):
        raise RuntimeError("predict error")
    def predict(self, X):
        pass


def accuracy_metric(y, pred):
    return (y.values == pred.values).mean()

def dummy_metric(y, pred):
    return 0.5


@pytest.fixture
def sample_data():
    np.random.seed(42)
    n = 100
    return pd.DataFrame({
        'f1': np.random.randn(n),
        'f2': np.random.randn(n),
        'f3': np.random.randn(n),
        'target': np.random.randint(0, 2, n),
    })


@pytest.fixture
def exp(tmp_path, sample_data):
    e = Experimenter(
        data=sample_data,
        path=tmp_path / 'exp',
        sp=ShuffleSplit(n_splits=2, test_size=0.2, random_state=42),
    )
    return e


def _setup_stage(e):
    e.set_grp('scale', role='stage', processor=StandardScaler,
              method='transform', edges={'X': [(None, ['f1', 'f2', 'f3'])]})
    e.set_node('scaler', grp='scale')


def _setup_head(e):
    e.set_grp('model', role='head', processor=DecisionTreeClassifier,
              method='predict', edges={'X': [(None, ['f1', 'f2', 'f3'])],
                                        'y': [(None, 'target')]},
              params={'max_depth': 3, 'random_state': 42})
    e.set_node('dt', grp='model')


def _setup_full(e):
    _setup_stage(e)
    _setup_head(e)


class TestDataCache:
    def test_put_get(self):
        c = DataCache(maxsize=1024**3)
        data = np.array([1, 2, 3])
        c.put_data('node1', 'all', 0, data)
        result = c.get_data('node1', 'all', 0)
        assert np.array_equal(result, data)

    def test_get_missing(self):
        c = DataCache(maxsize=1024**3)
        assert c.get_data('no_exist', 'all', 0) is None

    def test_clear_nodes(self):
        c = DataCache(maxsize=1024**3)
        c.put_data('a', 'all', 0, np.array([1]))
        c.put_data('b', 'all', 0, np.array([2]))
        c.clear_nodes(['a'])
        assert c.get_data('a', 'all', 0) is None
        assert c.get_data('b', 'all', 0) is not None

    def test_clear(self):
        c = DataCache(maxsize=1024**3)
        c.put_data('a', 'all', 0, np.array([1]))
        c.clear()
        assert c.get_data('a', 'all', 0) is None


class TestExperimenterInit:
    def test_path_created(self, exp):
        assert exp.path.exists()

    def test_data_wrapped(self, exp):
        assert exp.data is not None

    def test_splits_created(self, exp):
        assert exp.get_n_splits() == 2
        assert len(exp.valid_idx_list) == 2

    def test_no_inner_split(self, exp):
        assert exp.get_n_splits_inner() == 1
        train_idx, valid_v_idx = exp.train_idx_list[0][0]
        assert valid_v_idx is None

    def test_with_inner_split(self, tmp_path, sample_data):
        e = Experimenter(
            data=sample_data,
            path=tmp_path / 'exp_inner',
            sp=ShuffleSplit(n_splits=2, test_size=0.2, random_state=42),
            sp_v=KFold(n_splits=3, shuffle=True, random_state=42),
        )
        assert e.get_n_splits() == 2
        assert e.get_n_splits_inner() == 3
        train_idx, valid_v_idx = e.train_idx_list[0][0]
        assert valid_v_idx is not None

    def test_create_path_exists(self, tmp_path, sample_data):
        path = tmp_path / 'existing'
        path.mkdir()
        with pytest.raises(RuntimeError):
            Experimenter.create(data=sample_data, path=path)

    def test_status_open(self, exp):
        assert exp.status == 'open'

    def test_pipeline_empty(self, exp):
        assert len(exp.pipeline.grps) == 0

    def test_data_key(self, tmp_path, sample_data):
        e = Experimenter(data=sample_data, path=tmp_path / 'dk', data_key='test_key')
        assert e.data_key == 'test_key'


class TestPipelineDelegation:
    def test_set_grp(self, exp):
        r = exp.set_grp('g1', role='stage')
        assert r['result'] == 'new'
        assert 'g1' in exp.pipeline.grps

    def test_set_node(self, exp):
        exp.set_grp('g1', role='stage', processor=StandardScaler,
                    method='transform', edges={'X': [(None, None)]})
        r = exp.set_node('n1', grp='g1')
        assert r['result'] == 'new'
        assert 'n1' in exp.pipeline.nodes

    def test_remove_node(self, exp):
        _setup_stage(exp)
        exp.remove_node('scaler')
        assert 'scaler' not in exp.pipeline.nodes

    def test_remove_grp(self, exp):
        exp.set_grp('g1', role='stage')
        exp.remove_grp('g1')
        assert 'g1' not in exp.pipeline.grps

    def test_rename_grp(self, exp):
        exp.set_grp('old', role='stage')
        exp.rename_grp('old', 'new')
        assert 'new' in exp.pipeline.grps
        assert 'old' not in exp.pipeline.grps

    def test_closed_status_blocks_modification(self, exp):
        exp.close()
        with pytest.raises(RuntimeError):
            exp.set_grp('g1', role='stage')


class TestBuild:
    def test_build_stage(self, exp):
        _setup_stage(exp)
        exp.build()
        assert 'scaler' in exp.node_objs
        assert exp.node_objs['scaler'].status == 'built'

    def test_build_skips_built(self, exp):
        _setup_stage(exp)
        exp.build()
        n_objs_before = len(exp.node_objs)
        exp.build()
        assert len(exp.node_objs) == n_objs_before

    def test_build_error_continues(self, exp):
        exp.set_grp('good', role='stage', processor=StandardScaler,
                    method='transform', edges={'X': [(None, ['f1'])]})
        exp.set_node('good_node', grp='good')
        exp.set_grp('bad', role='stage', processor=BadProcessor,
                    method='transform', edges={'X': [(None, ['f2'])]})
        exp.set_node('bad_node', grp='bad')
        exp.build()
        assert exp.node_objs['good_node'].status == 'built'
        assert exp.node_objs['bad_node'].status == 'error'

    def test_build_error_dict(self, exp):
        exp.set_grp('err', role='stage', processor=ErrorProcessor,
                    method='transform', edges={'X': [(None, ['f1'])]})
        exp.set_node('err_node', grp='err')
        exp.build()
        err = exp.node_objs['err_node'].error
        assert err['type'] == 'TypeError'
        assert 'test error msg' in err['message']
        assert 'traceback' in err
        assert 'fold' in err


class TestExp:
    def test_exp_head(self, exp):
        _setup_full(exp)
        exp.build()
        exp.exp()
        assert 'dt' in exp.node_objs
        assert exp.node_objs['dt'].status == 'built'

    def test_exp_skips_built(self, exp):
        _setup_full(exp)
        exp.build()
        exp.exp()
        objs_before = dict(exp.node_objs)
        exp.exp()
        assert exp.node_objs['dt'] is objs_before['dt']

    def test_exp_error(self, exp):
        exp.set_grp('bad_model', role='head', processor=BadPredictor,
                    method='predict', edges={'X': [(None, ['f1'])],
                                              'y': [(None, 'target')]})
        exp.set_node('bad_dt', grp='bad_model')
        exp.exp()
        assert exp.node_objs['bad_dt'].status == 'error'

    def test_exp_with_collector(self, exp):
        _setup_full(exp)
        exp.build()
        mc = MetricCollector(
            'acc', Connector(),
            output_var=None,
            metric_func=accuracy_metric,
        )
        exp.add_collector(mc)
        exp.exp()
        assert mc.has('dt')


class TestCollectorManagement:
    def test_add_collector(self, exp):
        mc = MetricCollector('acc', Connector(),
                            output_var=None,
                            metric_func=dummy_metric)
        exp.add_collector(mc)
        assert 'acc' in exp.collectors
        assert mc.path is not None

    def test_add_collector_skip(self, exp):
        mc1 = MetricCollector('acc', Connector(),
                             output_var=None,
                             metric_func=dummy_metric)
        mc2 = MetricCollector('acc', Connector(),
                             output_var=None,
                             metric_func=dummy_metric)
        exp.add_collector(mc1)
        result = exp.add_collector(mc2, exist='skip')
        assert result is mc1

    def test_add_collector_error(self, exp):
        mc1 = MetricCollector('acc', Connector(),
                             output_var=None,
                             metric_func=dummy_metric)
        exp.add_collector(mc1)
        mc2 = MetricCollector('acc', Connector(),
                             output_var=None,
                             metric_func=dummy_metric)
        with pytest.raises(RuntimeError):
            exp.add_collector(mc2, exist='error')


class TestResetNodes:
    def test_reset_clears_node_objs(self, exp):
        _setup_stage(exp)
        exp.build()
        assert 'scaler' in exp.node_objs
        exp.reset_nodes(['scaler'])
        assert 'scaler' not in exp.node_objs

    def test_reset_clears_cache(self, exp):
        _setup_stage(exp)
        exp.build()
        exp.cache.put_data('scaler', 'all', 0, np.array([1]))
        exp.reset_nodes(['scaler'])
        assert exp.cache.get_data('scaler', 'all', 0) is None

    def test_set_node_replace_resets(self, exp):
        _setup_stage(exp)
        exp.build()
        assert 'scaler' in exp.node_objs
        exp.set_node('scaler', grp='scale', exist='replace')
        assert 'scaler' not in exp.node_objs


class TestStateManagement:
    def test_open_close(self, exp):
        exp.close()
        assert exp.status == 'close'
        exp.open()
        assert exp.status == 'open'

    def test_finalize_head(self, exp):
        _setup_full(exp)
        exp.build()
        exp.exp()
        exp.finalize(['dt'])
        assert exp.node_objs['dt'].status == 'finalized'

    def test_reinitialize(self, exp):
        _setup_full(exp)
        exp.build()
        exp.exp()
        exp.finalize(['dt'])
        exp.reinitialize(['dt'])
        assert 'dt' not in exp.node_objs


class TestSaveLoad:
    def test_save_creates_file(self, exp):
        _setup_stage(exp)
        assert (exp.path / '__exp.pkl').exists()

    def test_load_restores(self, exp, sample_data):
        _setup_full(exp)
        exp.build()
        exp.exp()
        path = exp.path

        loaded = Experimenter.load(path, sample_data)
        assert set(loaded.pipeline.grps.keys()) == set(exp.pipeline.grps.keys())
        assert 'scaler' in loaded.node_objs
        assert loaded.node_objs['scaler'].status == 'built'
        assert 'dt' in loaded.node_objs

    def test_load_data_key_mismatch(self, tmp_path, sample_data):
        e = Experimenter(data=sample_data, path=tmp_path / 'dk',
                        data_key='key_a')
        with pytest.raises(ValueError, match='data_key'):
            Experimenter.load(tmp_path / 'dk', sample_data, data_key='key_b')

    def test_load_preserves_splits(self, exp, sample_data):
        _setup_stage(exp)
        path = exp.path
        loaded = Experimenter.load(path, sample_data)
        assert loaded.get_n_splits() == exp.get_n_splits()
        assert loaded.get_n_splits_inner() == exp.get_n_splits_inner()


class TestPaths:
    def test_get_grp_path(self, exp):
        exp.set_grp('g1', role='stage')
        path = exp.get_grp_path('g1')
        assert path == exp.path / 'g1'

    def test_get_grp_path_nested(self, exp):
        exp.set_grp('parent', role='stage')
        exp.set_grp('child', role='stage', parent='parent')
        path = exp.get_grp_path('child')
        assert path == exp.path / 'parent' / 'child'

    def test_get_node_path(self, exp):
        _setup_stage(exp)
        path = exp.get_node_path('scaler')
        assert path == exp.path / 'scale' / 'scaler'


class TestExpObj:
    def test_stage_obj_lifecycle(self, tmp_path):
        path = tmp_path / 'stage_node'
        obj = StageObj(path)
        assert obj.status is None
        obj.start_build()
        assert path.exists()
        obj.end_build()
        assert obj.status == 'built'

    def test_stage_obj_start_exp_finalize_rejected(self, tmp_path):
        path = tmp_path / 'stage_node'
        obj = StageObj(path)
        with pytest.raises(ValueError):
            obj.start_exp(finalize=True)

    def test_stage_obj_exp_requires_built(self, tmp_path):
        path = tmp_path / 'stage_node'
        obj = StageObj(path)
        with pytest.raises(RuntimeError):
            list(obj.exp_idx(0, {}, iter([]), None))

    def test_stage_obj_exp_finalized_rejected(self, tmp_path):
        path = tmp_path / 'stage_node'
        obj = StageObj(path)
        obj.start_build()
        obj.end_build()
        obj.finalize()
        with pytest.raises(RuntimeError):
            list(obj.exp_idx(0, {}, iter([]), None))

    def test_stage_obj_end_exp(self, tmp_path):
        path = tmp_path / 'stage_node'
        obj = StageObj(path)
        obj.start_build()
        obj.end_build()
        obj.start_exp()
        obj.end_exp()
        assert obj.status == 'built'

    def test_stage_obj_finalize(self, tmp_path):
        path = tmp_path / 'stage_node'
        obj = StageObj(path)
        obj.start_build()
        obj.end_build()
        obj.finalize()
        assert obj.status == 'finalized'
        assert not path.exists()

    def test_head_obj_lifecycle(self, tmp_path):
        path = tmp_path / 'head_node'
        obj = HeadObj(path)
        assert obj.status is None
        obj.start_exp()
        assert path.exists()
        obj.end_exp()
        assert obj.status == 'built'

    def test_head_obj_finalize_after_exp(self, tmp_path):
        path = tmp_path / 'head_node'
        obj = HeadObj(path)
        obj.start_exp(finalize=True)
        obj.end_exp()
        assert obj.status == 'finalized'

    def test_head_obj_get_objs_requires_built(self, tmp_path):
        path = tmp_path / 'head_node'
        obj = HeadObj(path)
        with pytest.raises(RuntimeError):
            list(obj.get_objs(0))

    def test_stage_obj_load(self, tmp_path):
        path = tmp_path / 'stage_node'
        obj = StageObj(path)
        obj.load()
        assert obj.status == 'finalized'

    def test_head_obj_load(self, tmp_path):
        path = tmp_path / 'head_node'
        obj = HeadObj(path)
        obj.load()
        assert obj.status == 'finalized'



"""
Base adapter class for ML model frameworks
"""

from abc import ABC, abstractmethod


class ModelAdapter(ABC):
    result_objs = {}
    """Abstract base class for model adapters

    각 머신러닝 프레임워크별로 eval_set 처리 방식이 다르므로,
    이를 통일된 인터페이스로 추상화합니다.
    """

    def __init__(self, eval_mode='both', verbose=0.1):
        """Adapter 초기화

        Args:
            eval_mode (str): Evaluation mode
                - 'none' or None: eval_set 없이
                - 'valid': validation set만 전달
                - 'both': train + validation set 전달
            verbose: Verbose 설정
                - 0: 출력 안함
                - 0 < verbose < 1: 전체 진행률을 % 단위로 표시 (주기: verbose * 100%)
                  예: 0.1이면 10%마다, 0.05면 5%마다
                - verbose >= 1: iteration 단위로 표시 (매 verbose번째 iteration)
                  예: 1이면 매 iteration마다, 10이면 매 10 iteration마다
        """
        self.eval_mode = eval_mode
        self.verbose = verbose

    def get_fit_params(self, data_dict, X, y=None, params=None, logger=None):
        """모델의 fit()에 전달할 파라미터를 구성

        Args:
            data_dict: {key: (train, train_v), ...} 형태의 데이터 딕셔너리
            X: 입력 데이터의 key
            y: 타겟 데이터의 key (Optional, default=None)
            params (dict): Processor에서 전달된 추가 파라미터 (Optional, default=None)
            logger: Logger 인스턴스

        Returns:
            dict: fit()에 unpacking으로 전달할 파라미터
                  예: model.fit(**fit_params)
        """
        return {}

    def get_params(self, params, logger = None):
        """모델 생성자에 전달할 파라미터를 조정

        Args:
            params (dict): 원본 파라미터

        Returns:
            dict: 조정된 파라미터
        """
        return params
